## Sample from N(0,Q^-1):
## P*Q*Pt=L*Lt  =>  Q=Pt*L*Lt*P  =>  Q^-1=Pt*Lt^-1*L^-1*P
## So Pt*Lt^-1*u has the correct distribution where u~N(0,I)
## If n>1 then n independent columns are sampled.
GMRFsample <- function(Q,n=1,...){
  Q <- as(Q,"dsCMatrix")
  L <- Cholesky(Q,LDL=FALSE,...)
  m <- nrow(Q)
  u <- matrix(rnorm(m*n),m)
  x <- solve(L,solve(L,u,system="Lt"),system="Pt")
  as.matrix(x)
}

## Density of N(mu,Q^-1).
## If x is matrix the density is applied to each column of x.
GMRFdens <- function(x,Q,mu,log=TRUE,...){
  x <- as.matrix(x)
  ##n <- length(x)
  n <- nrow(x)
  if(!missing(mu)) x <- x-mu
  L0 <- Cholesky(Q,LDL=FALSE,...)
  L <- .Call("CHMfactor_to_sparse",L0,PACKAGE="Matrix")
  p <- L0@perm+1
  ##if(is.null(dim(x)))x <- x[p] else x <- x[p,]
  x <- x[p,]
  ##y <- t(L)%*%x
  y <- as.matrix(t(L)%*%x)
  logdetQ <- 2*sum(log(diag(L)))
  ##ans <- .5*logdetQ-.5*crossprod(y)
  ans <- .5*logdetQ-.5*colSums(y*y)
  ans <- as.numeric(ans)-.5*n*log(2*pi)
  if(!log)ans <- exp(ans)
  ans
}

## Standardize N(mu,Q^-1)
## (Reverse of GMRFsample): x=Pt*Lt^-1*u  =>  u=Lt*P*x
## If x is matrix then each column of x is standardized.
GMRFstandard <- function(x,Q,mu,...){
  x <- as.matrix(x)
  if(!missing(mu)) x <- x-mu
  L0 <- Cholesky(Q,LDL=FALSE,...)
  L <- .Call("CHMfactor_to_sparse",L0,PACKAGE="Matrix")
  p <- L0@perm+1
  ##x <- x[p]
  x <- x[p,]
  y <- t(L)%*%x
  as.matrix(y)
}

## Find precision of A*x with x~N(0,Q^-1).
## (A is a dimension reducing mapping - usually constructed as design-matrix
## corresponding to factor.)
## V(Ax)=A*VX*At=A*Pt*Lt^-1*L^-1*P*At=crossprod(L^-1*P*At)
GMRFreduce <- function(x,Q,At,...){
  L <- Cholesky(Q,LDL=FALSE,...)
  tmp <- solve(L,solve(L,At,system="P"),system="L")
  V <- crossprod(tmp)
  list(x=t(At)%*%x,
       Q=solve(V))
}

## Extend precision Q such that x|beta ~ N(A*beta,Q^-1) where beta~N(0,R^-1)
## If R is missing interpret as R=0 - diffuse prior.
GMRFextend <- function(Q,A,R){
  V <- -Q %*% A
  AtQA <- -crossprod(A, V)
  if (!missing(R))AtQA <- AtQA + R
  block2x2(Q, t(V), V, AtQA)
}

## Find marginal precision
GMRFmarginal <- function(Q,i,...){
  ## This works if i is "positive integer", "logical", "negative integer".
  ind <- 1:nrow(Q)
  i1 <- (ind)[i]
  i0 <- setdiff(ind,i1)
  if(length(i0)==0)return(Q)
  
  ## Find marginal prec of x[i1]
  Q0 <- as(Q[i0,i0,drop=FALSE],"dsCMatrix")
  L0 <- Cholesky(Q0,...)
  ans <- Q[i1,i1,drop=FALSE]-Q[i1,i0,drop=FALSE]%*%solve(Q0,Q[i0,i1,drop=FALSE])
  as(ans,class(Q))
}

## Conditional expectation of GMRF.
## Missing entries (NA) of x are replaced by conditional expectation.
GMRFpredict <- function(Q,x,sd=FALSE,...){
  x <- as.vector(x)
  if(length(x)!=nrow(Q))stop("Length of x must match dimension of Q")
  na <- is.na(x) ## Missing entries
  xB <- x[!na]
  QAA <- as(Q[na,na],"dsCMatrix")
  QAB <- Q[na,!na,drop=FALSE]
  L <- Cholesky(QAA, LDL = FALSE, ...)
  ##xA <- -solve(QAA,QAB%*%xB)
  xA <- -as.vector( solve(L,QAB%*%xB,system="A") )
  x[na] <- xA
  if(sd){
    sd <- rep(0,length(x))
    invQAA <- solveSubset(L)
    sd[na] <- sqrt(diag(invQAA))
    attr(x,"sd") <- sd
  }
  x
}

## Conditional expectation of LOG-GMRF.
## Missing entries (NA) of y are replaced by conditional expectation.
LGMRFpredict <- function(Q,y,...){ ## TODO: set sd-attribute to result
  x <- log(y)
  z <- GMRFpredict(Q,x,sd=TRUE,...)
  exp(z+.5*attr(z,"sd"))
}


##' Prediction for GMRF with separable covariance structure on d-dim lattice.
##'
##' Calculates conditional expectations for GMRF with separable covariance structure on d-dim lattice.
##' Is highly efficient utilizing both sparseness and separability.
##' 
##' @param Qlist List of precision matrices acting in each of the \code{length(dim(a))} dimensions.
##' @param a Array repesenting the lattice with NAs where no observation is available.
##' @param simulate Return a conditional sample instead of just predicting?
##' @param trace Trace the applied lattice expansions.
##' @return Updated array \code{a} with NAs replaced by predictions (or conditional simulations).
sepGMRFpredict <- function(Qlist,a,simulate=FALSE,trace=!simulate){
  ## helper: Matrix-vector operations for kronecker product. 
  ## Equivalent to "do.call(kronecker,rev(li)) %*% x"
  matvec <- function(li,x,FUN="%*%"){
    ## li <- rev(li)          <---- Kronecker(Q1,Q2) has Q2 fastest running !!
    FUN <- match.fun(FUN)
    n <- sapply(li,ncol)
    dim(x) <- n
    rotate <- function(x)aperm(x,c(seq.int(length=length(dim(x)))[-1],1L))
    for(i in seq(along=n)){
      y <- as.matrix( FUN(li[[i]],matrix(x,nrow(x))) )
      dim(y) <- c(nrow(y),dim(x)[-1])
      x <- rotate(y)
    }
    as.vector(x)
  }
  .cat <- function(...)if(trace)cat(...)
  Dim <- dim(a)
  ok <- identical(sapply(Qlist,nrow),dim(a))
  if(!ok)stop("Dimensions of precision matrices (",
              paste(sapply(Qlist,nrow),collapse=",") ,
              ") must match array dimension (",
              paste(dim(a),collapse=","),
              ")")
  nna <- !is.na(a)
  if(length(Dim)>1){
    numberInSlice <- lapply(seq(Dim),function(i) rowSums(aperm(nna ,c(i,seq(Dim)[-i]) )) )
  } else numberInSlice <- list(as.vector(nna))
  subDim <- sapply(numberInSlice,function(x)sum(x!=0))
  ok.lattice <- prod(subDim)==sum(nna)
  if(!ok.lattice).cat("The observation locations do not constitute a sub _lattice_\n")
  indx <- lapply(numberInSlice,function(x)x!=0)
  ## helpers:
  ## - extract a sub-lattice
  ## - modify a sub-lattice
  slice <- function(a,indx)do.call("[",c(list(a),indx,list(drop=FALSE)))
  "slice<-" <- function(x,indx,value)do.call("[<-",c(list(a),indx,list(value)))
  ## Initialize - project all dimensions to sub lattice
  li <- Map(GMRFmarginal,Qlist,indx)
  ## Observations not lattice --> Predict to sub-lattice first
  if(!ok.lattice){
    .cat("Predicting from observations to sub-lattice first.\n")
    asub <- slice(nna,indx) ## observation indicator restricted to smallest sub lattice.
    count <- lapply(seq(Dim),function(i) colSums(aperm(asub ,c(i,seq(Dim)[-i]) )) )
    mincount <- sapply(count,function(x)min(x[x>0]))
    fullRepresented <- mincount==subDim
    ## Keep full represented dimensions - collapse the others
    Qkeep <- li[fullRepresented]
    Qcollapse <- li[!fullRepresented]
    perm <- c(which(!fullRepresented), which(fullRepresented)  )
    Qkron <- as(t(1),"dgCMatrix");for(y in Qcollapse)Qkron <- y%x%Qkron
    ## Call *this* method to predict obs -> sub-lattice
    asub <- slice(a,indx)
    asub.perm <- aperm(asub,perm)
    olddim <- subDim[perm]
    newdim <- c(prod(subDim[!fullRepresented]),subDim[fullRepresented])
    dim(asub.perm) <- newdim
    .cat("Sub-lattice dimension:",paste(newdim,collapse=","),"\n")
    res <- sepGMRFpredict( c(list(Qkron),Qkeep), asub.perm, simulate=simulate, trace=trace )
    dim(res) <- olddim
    ## Permute result back to dimension of "asub" - and insert in large lattice
    iperm <- perm; iperm[perm] <- seq(perm)
    res <- aperm(res,iperm)
    slice(a,indx) <- res
    ## Now a has the required form - and we can call *this* method
    return( sepGMRFpredict( Qlist, a, simulate=simulate, trace=trace ) )
  }
  
  ## - Loop over dimensions to expand
  LoopDim <- which(!sapply(indx,all))
  for(i in LoopDim){
    ## Info
    dim1 <- sapply(indx,sum)
    dim2 <- dim1; dim2[i] <- Dim[i]
    txt <- paste(paste(dim1,collapse=","), paste(dim2,collapse=","), sep=" -> ")
    .cat("Lattice expansion: ",txt,"\n")
    ## Insert full matrix in dimension we are expanding
    li[[i]] <- Qlist[[i]]
    ## Sub-lattices
    indx0 <- indx                            ## Here we have observations
    indx1 <- indx; indx1[[i]] <- !indx1[[i]] ## Predict these entries
    ## Precision matrices relative to sub-lattices
    i0 <- indx0[[i]]
    i1 <- indx1[[i]]
    Q11 <- li; Q11[[i]] <- Q11[[i]][i1,i1,drop=FALSE]
    Q10 <- li; Q10[[i]] <- Q10[[i]][i1,i0,drop=FALSE]
    ## Formulas for prediction
    tmp1 <- matvec(Q10, slice(a,indx0))
    tmp2 <- -matvec(Q11, tmp1, solve)
    ## Add sample variation if requested
    if(simulate){
      ## Conditional precision of prediction-lattice given observations is Q11
      linfun <- function(Q,u){ ## linfun generates N(0,Q^-1) variates based on u~N(0,I)
        L <- Cholesky(Q, LDL = FALSE)
        solve(L, solve(L, u, system = "Lt"), system = "Pt")
      }
      u <- rnorm(prod(sapply(indx1,sum)))
      tmp2 <- tmp2+matvec(Q11,u,linfun)
    }
    ## insert result in array
    slice(a,indx1) <- tmp2
    ## Update lattice for which we have observations
    indx[[i]][] <- TRUE
  }
  return(a)
}
