#############################################################################
## CLASS: "covStruct"
##
## Return value of "fun" must be a list or an environment containing at least
## "precision" and "Dprecision"
#############################################################################
setClass("covStruct",representation(fun="function",
                                    parameters="numeric",
                                    name="character"))
setMethod("show","covStruct",
          function(object){
            cat("Covariance stucture:\n")
            cat(object@name,"\n\n")
            cat("Parameters:\n")
            print(object@parameters)
          })

## Calulate precision or derivatives for given parameters.
evaluate <- function(x,y=x@parameters){
  .lgc_derivmode(FALSE)
  do.call(x@fun,as.list(y))$precision
}
Devaluate <- function(x,y=x@parameters){
  .lgc_derivmode(TRUE)
  do.call(x@fun,as.list(y))$Dprecision
}

setMethod("image",signature(x="covStruct"),
          function(x,...)
          callGeneric( evaluate(x) ))
setMethod("plot",signature(x="covStruct"),
          function(x,y,...){
            par(mfrow=c(1,2))
            local({
              plot(dist,abs(solve(dense.covariance)),
                   log="y",ylab="Precision")
              plot(dist,abs(dense.covariance),
                   ylab="Covariance")
            },x@fun(...))
          })
setMethod("kronecker",signature(X="covStruct",Y="covStruct"),
          function(X,Y,FUN = "*", make.dimnames = FALSE, ...){
            force(X); force(Y)
            if(FUN!="*")stop("Not supported!")
            fun <- function(...){
              Xenv <- X@fun(...)
              Yenv <- Y@fun(...)
              if(exists("covariance",Xenv,inherits=FALSE)&
                 exists("covariance",Yenv,inherits=FALSE)){
              delayedAssign("covariance",new("tensor",matrices=
                                             list(Xenv$covariance,
                                                  Yenv$covariance))
                            )}
              delayedAssign("Dcovariance",
                            c.list(
                              lapply(Xenv$Dcovariance,
                                     function(d)new("tensor",matrices=
                                                    list(d,Yenv$covariance))),
                              lapply(Yenv$Dcovariance,
                                     function(d)new("tensor",matrices=
                                                    list(Xenv$covariance,d)))
                              )
                            )
              delayedAssign("precision",Xenv$precision %x% Yenv$precision)
              delayedAssign("Dprecision",
                            c.list(
                              lapply(Xenv$Dprecision,
                                     function(d)d%x%Yenv$precision),
                              lapply(Yenv$Dprecision,
                                     function(d)Xenv$precision%x%d)
                              ))
              environment()
            }
            parameters <- c(X@parameters,Y@parameters)
            ##formals(fun) <- c(as.list(parameters),formals(fun))
            new("covStruct",fun=fun,name=paste(X@name,Y@name,sep="  %x%  "),
                parameters=parameters)
          })



## cov-structures are usually constructed with function "covStruct"
## "dist" can either be a distance matrix or a list or env of distance matrices.
covStruct <- function(cov=exp(loga)*exp(-exp(logb)*dist)+exp(logc)*IdMat,
                      parameters=c(loga=0,logb=0,logc=0),
                      dist,
                      IdMat=diag(nrow(dist)),
                      radius,
                      neighbors,
                      pattern=
                      if(!missing(radius))as(dist<radius,"dgCMatrix")
                      else if(!missing(neighbors))nnb(dist,neighbors)
                      else 1,
                      ...){
  force(dist); force(pattern)
  cov <- substitute(cov)
  Dcov <- deriv(cov,names(parameters))
  have.pattern <- !identical(pattern,1)
  ##dot.args <- list(...)
  if(is.list(dist) | is.environment(dist)){
    dot.args <- dist
  } else {
    dot.args <- list(dist=dist,IdMat=IdMat,...)
  }
  fun <- function(...,DIST.ENV=dot.args){
    ## FIXME: Avoid setting parent.env.
    ## Need the union of variables of the present env and DIST.ENV
    ## The problem is that DIST.ENV may be constructed by e.g "dist.all"
    ## and therefore have a different env.
    if(is.environment(DIST.ENV))parent.env(DIST.ENV) <- environment()    
    ## Function should return:
    ## The "precision" and its derivative "Dprecision".
    ## "covariance" should be the inverse of "precision" and "Dcovariance"
    ## its derivative. Note that the covariance is only needed on the
    ## sparstity pattern given by the precision matrix i.e. it is sufficient
    ## to return covariance*pattern.
    ## Furthermore: the Dcovariance is only needed on the diagonal pattern.
    asSym <- function(x).5*(x+t(x))
    delayedAssign("dense.covariance",eval(cov,DIST.ENV))
    delayedAssign("dense.precision",
                  asSym(chol.solve(dense.covariance))) ## chol.solve!!!
    delayedAssign("precision",
                  if(have.pattern)dense.precision*pattern
                  else            dense.precision)
    delayedAssign("covariance",
                  if(have.pattern)solve(as(precision,"matrix"))
                  else            dense.covariance)
    delayedAssign("gradient",attr(eval(Dcov,DIST.ENV),"gradient"))
    delayedAssign("Dcovariance",
                  lapply(1:length(parameters),
                         function(i){
                           sigmadot <- gradient[,i]
                           dim(sigmadot) <- dim(covariance)
                           if(!have.pattern)return(sigmadot)
                           tmp <- (dense.precision%*%sigmadot
                                   %*%dense.precision)*pattern
                           as(covariance %*% tmp %*% covariance,"matrix")
                         })
                  )
    delayedAssign("Dprecision",
                  lapply(Dcovariance,
                         function(Dcov){
                           if(have.pattern)
                             -as(precision%*%Dcov%*%precision,"matrix")*pattern
                           else -precision%*%Dcov%*%precision
                         })
                  )

    environment()
  }
  formals(fun) <- c(as.list(parameters),formals(fun))
  new("covStruct",fun=fun,parameters=parameters,name=deparse(cov))
}

## Identity covariance structure of dimension n
covIdent <- function(n=length(x), x = rep.int(1, n), dense=FALSE){
  pattern <- as(.symDiagonal(n,x),"dgCMatrix")
  if(dense){
    tmp <- as(pattern,"dgeMatrix")
    pattern <- as(spMatrix(nrow(tmp),ncol(tmp),x=tmp@x,i=row(tmp),j=col(tmp)),"dgCMatrix")
  }
  fun <- function(...){
    precision <- pattern
    Dprecision <- Dcovariance <- list()
    environment()
  }
  new("covStruct",fun=fun,parameters=numeric(0),
      name=paste("covIdent(",")",sep=as.character(n)))
}  

## Stationary AR(k) structure on the line
## x[n]=phi[1]*x[n-1]+...+phi[k]*x[n-k]+eps[n]
## kappa chosen such that V(X_i)=1 so solve(Q) is correlation matrix.
## ------------------ Explanation ----------------------------------
## The mean-zero stationary AR(k) process with eps[n]~N(0,1) has density
## phi(x) ~ exp(-.5*x*t(D)*D*x) * mu(x_1,...,x_k)
## where D has dimension n by n+k and mu is the initial distribution.
## The resulting precision is Q=t(D)*D+Q0 where Q0 is the precision of
## the initial distribution - so Q0[1:k,1:k] are the only non-zeros in
## Q0. A necessary condition for x to be stationary is that the distribution
## of "x" is the same as "rev(x)". In terms of the precision this means that
## Q[1:n,1:n]=Q[n:1,n:1]. This identity uniquely determines the non-zeros
## Q0[1:k,1:k] of Q0. In fact Q0[1:k,1:k] is given by the "reversed"
## lower-rigth corner of t(D)*D minus the upper-left corner of t(D)*D.
## The resulting precision is easily found by "overwriting" the upper-left
## corner of t(D)*D with the reversed lower-right corner of t(D)*D.
## Finally we can scale Q such that solve(Q) is a correlation matrix. The
## scaling factor is easily found from the initial distribution:
##   solve(Q0[1:k,1:k])[1,1]
## -----------------------------------------------------------------
## corr: If corr=TRUE then scale with kappa to make solve(Q) a correlation matrix.
stat.ARk.prec <- function (phi = c(0.5, -5) * 0.1, n = 10, gradient = FALSE,
                           forcePD = TRUE, corr=TRUE) 
{
  ## Check parameters are ok
  require(polynom)
  coef <- c(rev(-phi),1)
  ok <- all(Mod(solve(polynomial(coef)))<1)
  if(!ok & forcePD)stop("Invalid parameter in stationary AR(k) precision")  
  ## Increment matrix D (ignoring the prior)
  ## Note: almost linear function of phi
  order <- length(phi)
  Dmat <- function(phi,one=1){
    n <- n-order
    x <- rep( rev(c( one ,-phi)) ,n)
    ans <- spMatrix(n,n+order,rep(1:n,each=order+1),outer(1:(order+1),0:(n-1),"+"),x)
    as(ans,"dgCMatrix")
  }
  ## D^T * D - with "reflection trick" to get stationarity
  ## Note: quadratic function of D
  Qmat <- function(D,D2=D){
    Q <- crossprod(D,D2) ## dgCMatrix
    ntail <- order^2  +  order*(order+1)/2
    Q@x[1:ntail] <- rev(tail(Q@x,ntail))
    Q
  }
  ## First row of Q^-1
  cov <- function(Q){
    M <- as(as(spMatrix(n,n,j=rep(n,n),i=1:n,x=rep(0,n)),"dgCMatrix"),"dsCMatrix")
    Q <- as(M+Q,"dsCMatrix")
    L <- Cholesky(Q,perm=FALSE)
    rev(solveSubset(L)[,n])
  }  
  D <- Dmat(phi)
  Q <- Qmat(D)
  if(!gradient & !corr)return(Q)
  c <- cov(Q)
  kappa <- c[1]
  ans <- kappa*Q
  if(!gradient & corr)return(ans)
  ## Derivatives
  derivD <- lapply(1:order,function(i){phi <- 0*phi;phi[i] <- 1;Dmat(phi,one=0)})
  derivQ <- lapply(derivD,function(Ddot)Qmat(D,Ddot)+Qmat(Ddot,D))
  if(!corr)return(derivQ)
  derivkappa <- lapply(derivQ,function(Qdot)as.numeric(-t(c)%*%Qdot%*%c))
  lapply(1:order,function(i)derivkappa[[i]]*Q+kappa*derivQ[[i]])
}

covStructAR <- function(parameters=c(phi1=.05,phi2=-.5),
                        n=10,...){
  force(n);
  fun <- function(...){
    phi <- c(...)[names(parameters)]
    delayedAssign("precision",stat.ARk.prec(phi,n=n))
    delayedAssign("Dprecision",stat.ARk.prec(phi,n=n,gradient=TRUE))
    delayedAssign("covariance",solve(as(precision,"matrix")))
    delayedAssign("Dcovariance",
                  lapply(Dprecision,
                         function(Dprec){
                           -as(covariance%*%Dprec%*%covariance,"matrix")
                         })
                  )                  
    environment()
  }
  new("covStruct",fun=fun,parameters=parameters,
      name=paste("Stationary AR(",length(parameters),")",sep=""))
}

## --------- Constant covariance structure -------------
## To be used with cov a fixed "dgCMatrix"
cov2covStruct <- function(cov){
  fun <- function(...){
    covariance <- cov
    Dcovariance <- list()
    delayedAssign("precision",solve(covariance))
    delayedAssign("Dprecision",list())
    environment()
  }
  new("covStruct",parameters=numeric(0),fun=fun)
}

## ------------------- Change parameterization of covStruct -----
## samePattern: Are the sparsity-patterns of the derivatives identical ?
setMethod("update",signature("covStruct"),
          function(object,
                   expr=NULL,
                   parameters=object@parameters,
                   fix=NULL,
                   samePattern=FALSE){
            ## Check arguments
            if(!is.null(expr)){
              if(!is.expression(expr))stop("expr must be expression")
              if(is.null(names(expr))){
                names(expr) <- names(object@parameters)
              }
            }
            ## experimental:
            string2expr <- function(x)structure(do.call(expression,lapply(x,as.name)),
                                                names=x)
            expr0 <- string2expr(names(object@parameters)) ## Default: identity map
            expr0[names(expr)] <- expr ## Overwrite with user-map
            ## Lookup parameters of updated mapping
            ## -Parameters of new mapping is "all.vars(expr0)"
            ## -First look in user supplied parameters
            ## -Then look in the original object
            parameters <- c(parameters,object@parameters)[all.vars(expr0)]
            if(!is.null(fix)){
              if(is.character(fix))fix <- parameters[fix]
              fix <- as.list(fix)
              li <- lapply(1:length(expr0),function(i)do.call("substitute",
                                                              list(expr=expr0[[i]],
                                                                   env=fix)))
              expr0[] <- li
              remove <- names(parameters) %in% names(fix)
              parameters <- parameters[!remove]
            }
            expr <- expr0
            ## -- end
            ok <- identical(sort(all.vars(expr)),sort(names(parameters)))
            if(!ok)stop("parameters dont match expression")
            if(length(parameters)>0)
              Dexpr <- lapply(expr,deriv,names(parameters),fun=names(parameters))
            else Dexpr <- list()
            reparam <- function(par){
              names(par) <- names(parameters)
              ans <- sapply(expr,eval,as.list(par))
              names(ans) <- names(object@parameters)
              ans
            }
            Dreparam <- function(par){
              names(par) <- names(parameters)
              sapply <- function(x,f)do.call(cbind,lapply(x,f))
              sapply(Dexpr,function(f)attr(do.call(f,as.list(par)),"gradient")[1,])
            }
            fun <- function(...){
              par <- c(...)[names(parameters)]
              if(is.null(par))par <- as.numeric(par)
              covenv <- do.call(object@fun,as.list(reparam(par)))
              Dpar <- Dreparam(par)
              chainrule <- function(env,var){
                Q <- get(var,envir=env)
                ans <- lapply(seq(along=par),
                              function(i){
                                j <- which(Dpar[i,]!=0)
                                Reduce("+",Map("*",Q[j],as.list(Dpar[i,j])))
                              })
                rm(list=var,envir=env)
                ans
              }
              if(samePattern){ ## Use more efficient method
                chainrule <- function(env,var){
                  Q <- get(var,envir=env)
                  mat <- sapply(Q,function(x)x@x)
                  rm(list=var,envir=env)
                  mat <- mat%*%t(Dpar)
                  for(i in seq(along=par))Q[[i]]@x <- mat[,i]
                  Q
                }
              }    
              delayedAssign("precision",covenv$precision)
              delayedAssign("Dprecision",
                            chainrule(covenv,"Dprecision"))
              delayedAssign("covariance",solve(as(precision,"matrix")))
              delayedAssign("Dcovariance",chainrule(covenv,"Dcovariance"))
              environment()  
            }
            new("covStruct",fun=fun,parameters=parameters,
                ## FIXME: name=object@name ???
                name=paste("Updated",object@name) )
          })

## --------------------------------------
## Add nugget to covStruct.
addNugget <- function (C,parameters=c(prec=1), entries) 
{
  myDiag <- function(n)as(spMatrix(n,n,1:n,1:n,rep(1,n)),"dgCMatrix")
  extend <- function(Q,nugget){
    A <- as(spMatrix(length(entries),nrow(Q),1:length(entries),
                  entries,rep(1,length(entries))),"dgCMatrix")
    D <- myDiag(length(entries))
    N <- D*nugget
    GMRFextend(N,A,Q)
  }
  fun <- function(...) {
    env <- C@fun(...)
    nugget <- c(...)[names(parameters)]
    delayedAssign("precision", extend(env$precision,nugget))
    delayedAssign("Dprecision",
                  c(rmLapply(env, "Dprecision", extend, nugget=0),
                    list( extend( 0*env$precision ,nugget=1)))
                  )
    environment()
  }
  new("covStruct", fun = fun, parameters = c(C@parameters,parameters),
      name = c(C@name,"extended"))
}
## Utility fct
## Pattern corresponding to n-nearest neighbor
nnb <- function(dist,n=3){
  mat <- apply(dist,1,function(x)x<=sort(x)[n]+1e-12)
  as(mat|t(mat),"dgCMatrix")
}

## ------- Extend covariance structure:
## Get precision of field X with design matrix lookup AX.
##   [ AX ]
##   [ X  ]
## Where Prec(X) and A are input. Singular unless delta>0.
extendCovStruct <- function(C,A,delta=1e-10){
  extend <- function(Q,PD=FALSE){
    V <- -Q%*%A
    AtQA <- -crossprod(A,V)
    if(PD)diag(AtQA) <- diag(AtQA)+delta
    block2x2(Q,t(V),V,AtQA)
  }
  fun <- function(...){
    env <- C@fun(...)
    delayedAssign("precision",extend(env$precision,TRUE))
    delayedAssign("Dprecision",rmLapply(env,"Dprecision",extend))
    delayedAssign("covariance",chol.solve(precision))
    environment()
  }
  new("covStruct", fun = fun, parameters = C@parameters,
      name = c(C@name,"extended"))
}

## ------- Extend covariance structure:
## Get precision of
##   [ AX+Y ]
##   [ X    ]
## Where Prec(X), Prec(Y) and A are input
axpyCovStruct <- function (X, Y, A, delta = 1e-10) 
{
  force(A);force(X);force(Y)
  extend <- function(X,Y)GMRFextend(Q=Y,A=A,R=X)
  extendDX <- function(DX,YY)GMRFextend(Q=ZeroMatrix(dim(YY)),
                                       A=ZeroMatrix(dim(A)),
                                       R=DX)
  extendDY <- function(DY,XX)GMRFextend(Q=DY,
                                       A=A,
                                       R=ZeroMatrix(dim(XX)))
  fun <- function(...) {
    envX <- X@fun(...)
    envY <- Y@fun(...)
    delayedAssign("precision", extend(envX$precision,
                                      envY$precision) )
    delayedAssign("Dprecision",
                  c(
                    lapply(envX$Dprecision,extendDX,YY=envY$precision),
                    lapply(envY$Dprecision,extendDY,XX=envX$precision)
                    ) )
    environment()
  }
  new("covStruct", fun = fun, parameters = c(X@parameters,Y@parameters),
      name = c(X@name, "axpy", Y@name))
}

## ----------- Block diagonal covariance structure
bdiagCovStruct <- function(x,...){
  if(is.list(x))args <- x else args <- list(x,...)
  plist <- lapply(args,function(x)x@parameters)
  names(plist) <- NULL
  p <- unlist(plist)
  p <- p[!duplicated(names(p))]  ## Union of parameters
  lapply(plist,function(x)which(names(p)%in%names(x)))  
  fun <- function(...){
    envList <- lapply(args,function(x)x@fun(...))
    delayedAssign("precision", mybdiag(lapply(envList,function(x)x$precision)) )
    delayedAssign("dims", lapply(envList,function(x)dim(x$precision)) )
    delayedAssign("Ztemplate",Map(ZeroMatrix,dims))
    Df <- function(name){
      Z <- Ztemplate
      li <- lapply(plist,function(x)which(name==names(x)))
      for(i in 1:length(args)){        
        if(length(li[[i]])>0)Z[[i]] <- envList[[i]]$Dprecision[[li[[i]]]]
      }
      mybdiag(Z)
    }
    delayedAssign("Dprecision", lapply(names(p),Df) )
    environment()
  }
  new("covStruct", fun = fun, parameters = p,
      name = c("bdiag(", paste(sapply(args,function(x)x@name),collapse=", "  ), ")" ) )
  
}

## -------------------------------
scaleCovStruct <- function(C,scale){
  pnew <- rep(1,nlevels(scale))
  names(pnew) <- paste("scale",levels(scale),sep="")
  Dtemplate <- apply(modMat(scale),2,function(x)Diagonal(x=x))
  parameters <- c(C@parameters,pnew)
  fun <- function(...){
    pnew <- c(...)[names(pnew)]
    env <- C@fun(...)
    delayedAssign("D" , Diagonal(x = pnew[scale] )   )
    delayedAssign("DD" , Dtemplate )
    delayedAssign("precision", D %*% env$precision %*% D )
    delayedAssign("Dprecision1", lapply(env$Dprecision,function(x)D%*%x%*%D) )
    delayedAssign("Dprecision2", lapply(DD,function(x)x%*%env$precision%*%D +
                                        D%*%env$precision%*%x  ) )
    delayedAssign("Dprecision" , c(Dprecision1,Dprecision2) )
    environment()
  }
  new("covStruct", fun = fun, parameters = parameters,
      name = c("scaled(", C@name  , ")" ) )  
}

## ------- Parameter:
covParameter <- function(expr=exp(logscale),parameters=c(logscale=0)){
  cov <- substitute((x)*IdMat,list(x=substitute(expr)))
  args <- list(cov=cov,dist=t(0),parameters=parameters,radius=1)
  do.call(covStruct,args)
}
    
## ------- Check derivatives of covStruct
## * Allow Q="dsCMatrix" with Qdot="dgCMatrix"
## * Can pass argument "i" to check on pattern subset
checkCovStruct <- function (C,i) 
{
    require(numDeriv)
    V <- evaluate(C)
    if(missing(i))V@x[] <- rnorm(length(V@x))
    else {V@x[] <- 0;V@x[i] <- 1}
    h <- function(x) sum((V * x))
    f <- function(x) h(evaluate(C,x))
    Df <- function(x) sapply(Devaluate(C,x),h)
    ans <- Df(C@parameters) - grad(f, C@parameters)
    names(ans) <- names(C@parameters)
    ans
}

covDiag <- function (fac, parameters=rep(1,nlevels(fac)), samePattern=FALSE ) 
{
  names(parameters) <- levels(fac)
  n <- length(fac)
  ind <- 1:n
  spl <- split(ind,fac)
  myf <- function(i)as(spMatrix(n,n,i,i,rep(1,length(i))),"dgCMatrix")
  P <- lapply(spl,myf)
  if(samePattern)P <- lapply(P,function(x)as(.symDiagonal(nrow(x),diag(x)),"dgCMatrix"))
  fun <- function(...) {
    par <- c(...)[names(parameters)]
    x <- par[fac]
    precision <- as(.symDiagonal(n, x), "dgCMatrix")
    Dprecision <- P
    environment()
  }
    new("covStruct", fun = fun, parameters = parameters, name = paste("covDiag(", 
        ")", sep = as.character(n)))
}

## Log-transform parameters of covStruct.
setMethod("log","covStruct",function(x){
  p <- log(x@parameters)
  names(p) <- paste("log",names(x@parameters),sep="")
  expr <- as.expression(lapply(names(p),function(y)substitute(exp(x),list(x=as.name(y)))))
  names(expr) <- names(x@parameters)
  update(x,expr,parameters=p)
})

## Same as axpyCovStruct but now with "ZAX+Y" where Z is diagonal matrix with
## scaling parameters.
## Z is specified through factor "scale".
## [ Q       -QZA     ]
## [ -AtZQ   AtZQZA+R ]
axpyCovStruct2 <- function (X, Y, A, delta = 1e-10, scale, Z=log(covDiag(scale))) 
{
  force(A);force(X);force(Y);force(Z)
  extend <- function(X,Y,Z)GMRFextend(Q=Y,A=Z%*%A,R=X)
  extendDX <- function(DX,YY,ZZ)GMRFextend(Q=ZeroMatrix(dim(YY)),
                                           A=ZeroMatrix(dim(A)),
                                           R=DX)
  extendDY <- function(DY,XX,ZZ)GMRFextend(Q=DY,
                                           A=ZZ%*%A,
                                           R=ZeroMatrix(dim(XX)))
  extendDZ <- function(DZ,XX,YY,ZZ){
    ZA <- ZZ%*%A
    DZA <- DZ%*%A
    V <- -YY %*% DZA
    DAtQA <- -crossprod(ZA, V)
    block2x2(ZeroMatrix(dim(YY)), t(V), V, DAtQA + t(DAtQA))
  }
  fun <- function(...) {
    envX <- X@fun(...)
    envY <- Y@fun(...)
    envZ <- Z@fun(...)
    delayedAssign("precision", extend(envX$precision,
                                      envY$precision,
                                      envZ$precision) )
    delayedAssign("Dprecision",
                  c.list(
                    lapply(envX$Dprecision,extendDX,YY=envY$precision,ZZ=envZ$precision),
                    lapply(envY$Dprecision,extendDY,XX=envX$precision,ZZ=envZ$precision),
                    lapply(envZ$Dprecision,extendDZ,XX=envX$precision,YY=envY$precision,
                           ZZ=envZ$precision)
                    ) )
    environment()
  }
  new("covStruct", fun = fun, parameters = c(X@parameters,Y@parameters,Z@parameters),
      name = c(X@name, "axpy", Y@name))
}

## General corr matrix
covStructGeneral <- function(ch,basename,pdTransform=FALSE,...){
  n <- length(ch)
  combn <- combn(ch,2)
  D <- diag(length(ch))
  comb <- combn(seq(ch),2)
  E <- function(i,j){m <- matrix(0,n,n);m[i,j] <- m[j,i] <- 1;m}
  li <- Map(E,comb[1,],comb[2,])
  names(li) <- paste("e",comb[1,],comb[2,],sep=".")
  pn <- paste(basename,combn[1,],combn[2,],sep=".")
  pn2 <- paste("`",pn,"`",sep="")
  if(pdTransform){
    comb <- cbind(comb,t(cbind(1:n,1:n)))
    li <- Map(E,comb[1,],comb[2,])
    names(li) <- paste("e",comb[1,],comb[2,],sep=".")
    D[] <- 0
    U <- outer(ch,ch,function(x,y)paste(basename,x,y,sep="."))
    U[] <- paste("`",U,"`",sep="")
    diag(U) <- "1"
    U[lower.tri(U)] <- ""
    Ut <- t(U)
    inner.prod <- function(x,y){
      i <- (x!="")&(y!="")
      paste("(",paste(x[i],y[i],sep="*",collapse="+"),")")
    }
    pn2 <- apply(comb,2,function(ind)inner.prod(Ut[ind[1],], U[,ind[2]]))
    d1 <- apply(comb,2,function(ind)paste("sqrt",inner.prod(Ut[ind[1],], U[,ind[1]] )))
    d2 <- apply(comb,2,function(ind)paste("sqrt",inner.prod(Ut[ind[2],], U[,ind[2]] )))
    pn2 <- paste(pn2,d1,d2,sep="/")
  }  
  expr <- parse(text=paste(c(paste(pn2,names(li),sep="*"),"D"),collapse="+"))[[1]]
  li <- c(li,list(D=D))
  p <- rep(0,length(pn))
  names(p) <- pn
  args <- list(cov=expr,dist=li,parameters=p,pattern=as(D*0+1,"dgCMatrix"),...)
  do.call("covStruct",args)
}

## Project covStruct
covStructProj <- function(C,i,perm=TRUE,super=FALSE,dense=FALSE){
  fun <- function(...){
    envC <- C@fun(...)

    delayedAssign("precision",
                  {
                    Q <- envC$precision
                    ind <- 1:nrow(Q)
                    i1 <- (ind)[i]
                    i0 <- setdiff(ind, i1)
                    if (length(i0) == 0) Q else {
                      Q0 <- as(Q[i0, i0, drop = FALSE], "dsCMatrix")
                      L0 <- Cholesky(Q0, perm=perm,super=super)
                      A <- solve(Q0, Q[i0, i1, drop = FALSE])
                      ans <- Q[i1, i1, drop = FALSE] - Q[i1, i0, drop = FALSE] %*% 
                        A
                      as(ans, class(Q))
                    }
                  })
    ## I ~ i1
    ## A ~ i0 x i1
    delayedAssign("Dprecision",
                  {
                    force(precision)
                    DQ <- envC$Dprecision
                    if (length(i0) == 0) DQ else {
                      I <- .symDiagonal(length(i1))
                      B <- rbind2(-I,A)
                      if(dense)B <- as(B,"dgeMatrix")
                      p <- c(i1,i0)
                      p[p] <- seq(p) ## inverse perm
                      B <- B[p,]
                      ## Bt P DQ Pt B = (Pt B)t DQ (Pt B)
                      ## lapply(DQ,function(DQ)t(B)%*%(DQ[p,p]%*%B))
                      ans <- lapply(DQ,function(DQ)t(B)%*%(DQ%*%B))
                      if(dense)ans <- lapply(ans,function(x)as(x,"dgCMatrix"))
                      ans
                    }
                  })
    environment()
  }
  new("covStruct", fun = fun, parameters = C@parameters,
      name = c("Projected",C@name))  
  
}

## Analogue of "cov2cor"
## Fixme: Derivatives are inefficient!
covStruct2Cor <- function(C){
  fun <- function(...){
    envC <- C@fun(...)
    ## cov2cor(C)    = diag(C)^-.5 %*% C %*% diag(C)^-.5
    ## cov2cor(C)^-1 = diag(C)^.5  %*% Q %*% diag(C)^.5
    delayedAssign("precision",
                  {
                    Q <- envC$precision
                    L <- Cholesky(as(Q,"dsCMatrix"),perm=!TRUE,LDL=FALSE)
                    sd <- sqrt(diag(solveSubset(L)))
                    Diagonal(x=sd)%*%Q%*%Diagonal(x=sd)
                  })
    delayedAssign("Dprecision",
                  {
                    force(precision)
                    DQ <- envC$Dprecision
                    covariance <- solve(Q)
                    diagC.5 <- Diagonal(x=sqrt(diag(covariance)))
                    DdiagC.5 <- 
                      lapply(DQ,
                             function(Dprec){
                               DdiagC <- -diag(covariance%*%Dprec%*%covariance)
                               x <- .5*diag(covariance)^-.5 * DdiagC
                               Diagonal(x=x)
                             })
                    lapply(seq(DQ),function(i)
                           DdiagC.5[[i]] %*% Q       %*% diagC.5+
                           diagC.5       %*% DQ[[i]] %*% diagC.5+
                           diagC.5       %*% Q       %*% DdiagC.5[[i]]
                           )
                    
                  })
    environment()
  }
  new("covStruct", fun = fun, parameters = C@parameters,
      name = c("Cov2cor",C@name))  
}

## Non-linear fixed effects
covStructNonlin <- function(f,parameters,delta=1e-10,invdelta=1/delta){
  extend <- function(A){
    A <- as.matrix(A)
    Q <- .symDiagonal(length(A),invdelta)
    block2x2(Q,-invdelta*t(A),-invdelta*A,invdelta*crossprod(A)+delta)
  }
  Dextend <- function(DA,A){
    A <- as.matrix(A)
    DA <- as.matrix(DA)
    DQ <- ZeroMatrix(rep(length(A),2))
    block2x2(DQ,-invdelta*t(DA),-invdelta*DA,invdelta*2*crossprod(A,DA))
  }
  fun <- function(...){
    par <- c(...)[names(parameters)]
    A <- do.call("f",as.list(par))
    delayedAssign("DA",attr(A,"gradient"))
    delayedAssign("precision",extend(A))
    delayedAssign("Dprecision",lapply(split(DA,col(DA)),Dextend,A=A))
    environment()
  }
  new("covStruct", fun = fun, parameters = parameters,
      name = c("nonlin"))  
}

## Non linear fixed effects - another strategy
## Input: * Existing covariance stucture object.
## Output: * Modify by binding single column with non-linear effect
## Vector output f(x) is assumed to be mapped to data values which
## by convention are assumed to be indexed from first index of C.
covStructNonlinExtend <- function(C,f,parameters,delta=1e-10){
  delta <- as(matrix(delta),"dgCMatrix")
  extend <- function(A,Q)GMRFextend(Q=Q,A=A,R=delta)
  extendDQ <- function(DQ,A)GMRFextend(Q=DQ,A=A)
  extendDA <- function(DA,Q,A){
    Z <- ZeroMatrix(dim(Q))
    V <- -Q %*% DA
    AtQDA <- -crossprod(A, V)
    block2x2(Z, t(V), V, 2*AtQDA)
  }
  splitCol <- function(mat)lapply(1:ncol(mat),function(i)mat[,i])
  fun <- function(...){
    envC <- C@fun(...)    
    par <- c(...)[names(parameters)]
    A <- do.call("f",as.list(par))
    delayedAssign("DA",lapply( splitCol(attr(A,"gradient")) ,spvec))
    delayedAssign("Q",envC$precision)
    delayedAssign("DQ",envC$Dprecision)
    delayedAssign("spA",spvec(A))
    spvec <- function(x){
      n <- nrow(Q)
      lx <- length(x)
      new("dgCMatrix",x=as.numeric(x),i=as.integer(0:(lx-1)),
          p=as.integer(c(0,lx)),Dim=as.integer(c(n,1)))
    }
    delayedAssign("precision",extend(spvec(A),Q))
    delayedAssign("Dprecision",c.list( lapply(DQ,extendDQ,A=spA),
                                      lapply(DA,extendDA,Q=Q,A=spA) ))

    environment()
  }
  new("covStruct", fun = fun, parameters = c(C@parameters,parameters),
      name = c("nonlinExtend",C@name))  
  
}

## Obtain "pattern" through "triang" function.
## Note: Not rescaled to correlation matrix.
covStructGMRF <- function(pattern,parameters=c(delta=1)){
  pattern <- as(pattern,"dgCMatrix")
  pattern@x[] <- 1
  ##nb <- max(rowSums(pattern))-1 ## Number of neighbors
  nb <- as.numeric(rowSums(pattern))-1 ## Number of neighbors
  fun <- function(...){
    delta <- c(...)[names(parameters)]
    delayedAssign("precision",{
      pattern@x[] <- -1
      diag(pattern) <- nb+delta
      pattern
    })
    delayedAssign("Dprecision",{
      list({
        pattern@x[] <- 0
        diag(pattern) <- 1
        pattern
      })
    })
    environment()
  }
  #formals(fun) <- c(as.list(parameters), formals(fun))
  new("covStruct",fun=fun,parameters=parameters,
      name=paste("GMRF(",length(parameters),")",sep=""))
}

## ============= 1st order vector autoregressive process "VAR1" ============
## X_t = AX_{t-1} + eps_t, where eps_t ~ N(0,kappa^-1 * I)
## X_0 ~ N(0,PRIOR^-1)
##
## Case kappa=1. Density is (WITHOUT PRIOR)
## ----------------------------------------
##
## -log pi(x) = .5 sum_i (x_i - A x_{i-1} )' (x_i - A x_{i-1} )
##
## 2nd order derivative of i'th term:
## [ A'A  -A' ]
## [ -A    I  ]
##
## Total 2nd order derivative (e.g. with n=4): 
## [ A'A  -A'                      ]
## [ -A   I+A'A  -A'               ]
## [      -A     I+A'A  -A'        ]  =  D'D = Q
## [             -A     I+A'A  -A' ]
## [                    -A      I  ]
##
## where D=
##
## [ -A   I             ]
## [     -A   I         ]
## [         -A   I     ]
## [             -A   I ]
##
## =
##
## [ -A               0  ]    [  0   I             ]   
## [     -A           0  ] +  [  0       I         ]
## [         -A       0  ]    [  0           I     ]
## [             -A   0  ]    [  0               I ]
##
## ----------------------------------------
##
## SAME *WITH* PRIOR Q0=I-A'A (Inspired by steady-state precision in 1D-case)
## Total 2nd order derivative (e.g. with n=4):
## ("Overwrite" top-left block "A'A" with "I" )
## [  I   -A'                      ]
## [ -A   I+A'A  -A'               ]
## [      -A     I+A'A  -A'        ]
## [             -A     I+A'A  -A' ]
## [                    -A      I  ]
##
## Lower Cholesky:
## [  I                ]
## [ -A   I            ]
## [     -A  I         ]
## [         -A  I     ]
## [            -A  L  ]
##
## Where L*L' = I-A'A

## normal: Is the matrix A normal (A'A=AA') ?
## If not then the Lyapunov equation is needed to find steady-state
## variance.
blockAR1 <- function(A,
                     PRIOR = .symDiagonal(nrow(A[[1]]),delta),
                     n,kappa=1,delta=1e-6,simulate=FALSE,
                     gradient=FALSE,Adot,stationary=FALSE,
                     normal=FALSE){
  if(!is.list(A))A <- list(A)
  if(missing(n))n <- length(A) else n <- n-1
  A <- rep(A,length.out=n)
  dimA <- nrow(A[[1]])
  if(stationary){
    if(normal)PRIOR <- .symDiagonal(dimA)-crossprod(A[[1]])
    else PRIOR <- lyapunov(A[[1]])
  }
  if(simulate){
    x0 <- GMRFsample(PRIOR)
    fun <- function(x,A)as.vector(A%*%x)+rnorm(length(x),sd=sqrt(kappa)^-1)
    ans <- Reduce(fun,A,init=x0,acc=TRUE)
    return(unlist(ans))
  }
  D <- -mybdiag(A)
  I <- .symDiagonal(nrow(D))
  Z <- ZeroMatrix(c(nrow(D),ncol(A[[1]])))
  D <- cbind2(D,Z)+cbind2(Z,I)

  addPrior <- function(Q,PRIOR)Q+mybdiag(list(PRIOR,ZeroMatrix(dim(Q)-dim(PRIOR))))

  if(gradient){  ## assuming no parameters in the prior - e.g. flat prior.
                 ## And kappa=1
    Qdot <- function(Adot){
      if(!is.list(Adot))Adot <- list(Adot)
      Adot <- rep(Adot,length.out=n)
      Ddot <- cbind2( -mybdiag(Adot), Z )
      ##  D'D ----> Ddot'D + D'Ddot
      DdotD <- t(Ddot)%*%D
      ans <- DdotD + t(DdotD)
      if(stationary){
        if(normal)ans[1:dimA,1:dimA] <- 0
        else ans[1:dimA,1:dimA] <- ans[1:dimA,1:dimA]+
          lyapunov(A[[1]],list(Adot[[1]]))[[1]]
      }
      ans
    }
    ans <- lapply(Adot,Qdot)
    return(ans)
  }

  ##PRIOR <- mybdiag( list( PRIOR, ZeroMatrix(dim(ans)-dim(PRIOR))  ) )
  ##ans <- ans+PRIOR
  Q <- crossprod(D)
  ans <- kappa*addPrior(Q,PRIOR)
  ans
}
## Find stationary *precision* V^-1 for given A
## and possibly derivatives, where V=AVA'+I
lyapunov <- function(A,Adot=NULL){
  n <- nrow(A)
  I <- .symDiagonal(n)
  B <- I%x%I-A%x%A
  vecI <- as.vector(diag(n))
  V <- solve(B,vecI)
  dim(V) <- c(n,n)
  Q <- solve(V)
  Qdot <- function(Adot){
    rhs <- Adot %*% V %*% t(A)
    rhs <- rhs+t(rhs)
    Vdot <- solve(B,as.vector(rhs))
    dim(Vdot) <- c(n,n)
    as(-Q %*% Vdot %*% Q,"dgCMatrix")
  }
  if(is.null(Adot))as(Q,"dgCMatrix") else lapply(Adot,Qdot)
}

## CovStruct
covStructVAR1 <- function(operator,ntime,stationary=FALSE,delta=1e-6,
                          normal=FALSE,...){
  form <- formals(operator)
  form <- form[names(form)!="gradient"]
  parameters <- unlist(form)
  fun <- function(...){
    parms <- list(...)[names(parameters)]
    A <- do.call(operator,parms)
    precision <- blockAR1(A,n=ntime,kappa=1,stationary=stationary,delta=delta,
                          normal=normal)
    delayedAssign("Adot",do.call(operator,c(parms,gradient=TRUE)))
    delayedAssign("Dprecision",blockAR1(A,n=ntime,kappa=1,gradient=TRUE,Adot=Adot,stationary=stationary,delta=delta,
                                        normal=normal))
    environment()
  }
  new("covStruct", fun = fun, parameters = parameters,
      name = paste("VAR1(",length(parameters), ")", sep = ""))
}
