##' Formula interface to covStructs.
##' 
##' Specify a full Gaussian Markov Random Field via formulas.
##'
##' The present interface allow the construction of Gaussian Random Fields
##' of the form \eqn{Z=Z'+A \beta} where \eqn{Z'=B Y+\epsilon} and \eqn{Y=c(X_1,...,X_k}).
##' With this notation \eqn{Y} is the mean zero hidden random field which is a composition of
##' multiple hidden multivariate layers \eqn{X_1,...,X_k}. The design matrix \eqn{B} plays the role of
##' looking up the relevant contributions of \eqn{X_1,...,X_k} to each observation entry.
##' The construction is controlled by the arguments
##' \itemize{
##'   \item \code{formula} Generates the Individual processes \eqn{X_1,...,X_k} and the design matrix \eqn{B}.
##'   \item \code{fixef} Generates fixed effect design matrix \eqn{A}.
##'   \item \code{nugef} Generates the nugget effect \eqn{\epsilon}.
##' }
##' Currently defined syntax of the random field \code{formula}
##' \enumerate{
##' \item \code{ID(factor)} Defines IID process with variance one on the levels of
##'   \code{factor}.
##' \item \code{DIAG(factor)} Defines a process of independent variables with a
##'   variance parameter for each level of \code{factor}.
##' \item \code{AR1(factor)} Defines a stationary AR1 process with variance one on the
##'   levels of \code{factor}.
##' \item \code{AR2(factor)} Defines a stationary AR2 process with variance one on the
##'   levels of \code{factor}.
##' \item \code{contAR2(numeric)} Defines a variance 1 process with covariance structure
##' \eqn{\rho(t)=\exp(a*t/s)*(\cos(b*t/s)-a/b*\sin(b*t/s))}, \eqn{b=\sqrt{1+a^2}} (with derivative
##' augmentation to obtain exact sparseness of the precision matrix on any irregular grid).
##' The parameter \eqn{a} is a shape parameter in the interval \eqn{(-1,0)} and \eqn{s} is a
##' positive scale parameter. The structure solves a linear 2nd order ODE and can approximate most
##' powered exponential correlations. For instance the shape \eqn{a=-0.77} is close to a Gaussian
##' correlation.
##' \item \code{CORR(factor)} Defines a free un-structured correlation matrix on the
##'   levels of \code{factor}.
##' \item \code{GMRF1(gridFactor)} Defines a first order Gaussian Markov Random Field
##'   on the grid associated with \code{gridFactor}.
##'   Is neither stationary nor of variance one. The model is also known as "Besags proper model".
##' \item \code{COV(distFactor,form)} Defines a process in terms of the covariance
##'   specified through the formula \code{form}.
##'   Distance matrices (or increment matrices) are looked up in the
##'   level attributes of \code{distFactor}. 
##' \item \code{*} The notation \code{X*Y} generates a process on the product space
##'   with the multiplicative covariance structure.
##' \item \code{+} The notation \code{X+Y} generates independent processes X and Y
##'   and relates the observation to the sum.
##' \item \code{:} The notation \code{a:X} relates the observation to the process X
##'   scaled with a covariate a.
##' \item \code{PARM(name)} Generates a scale parameter that can be multiplied with a term.
##' \item \code{SCALE(X)} Multiplies different scale parameters on each component of
##'   the process \code{X} (should only be applied to reasonably low dimensional \code{X}).
##'   The aggregation of scale parameters can be controlled by \code{SCALE(X,fac)} where \code{fac}
##'   is a factor with length equal to the dimension of \code{X}. This will add \code{nlevels(fac)}
##'   scale parameters.
##' \item \code{PROJ(X)} Project the covariance structure by removing entries for which no
##' direct data-entry is associated. This does not change the model but only affects the internal
##' sparse representation and computation time. Note that projection generally destroys sparseness.
##' However it can sometimes improve performance on large separable fields by replacing e.g \code{X*Y*Z}
##' by \code{X*Y*PROJ(Z)}.
##' }
##' Note that factors with NAs can be used to indicate that there is no relation
##' between a given observation and one or more of the random field terms.
##' This may be useful in situations where multiple different data sources are modelled.
##'
##' @param formula Random effect formula (details below).
##' @param data Data.frame
##' @param fixef Fixed effect formula. NAs in model.matrix are replaced by zeros.
##' @param delta Tiny number. Precision of flat prior on fixed effects. 
##' @param invdelta Huge number. Precision of nugget effect when no nugget is desired.
##' @param nugef Nugget effect formula.
##' @param update.args List used to re-parameterize the covStruct.
##' @return Object of class covStruct with attributes used by \code{fitlgc}.
##' @seealso \code{\link{fitlgc}}
##' @examples
##' ##' === Separable correlation on grid with a nugget effect.
##' d <- expand.grid(x=factor(1:7),y=factor(1:8))
##' C <- covStructFormula(~AR1(x)*AR1(y)+AR2(y),nugef=~1,fixef=NULL,data=d)
##' ##' === Some methods
##' C@@parameters
##' simulate(C)
##' image(C)
covStructFormula <- function(formula,data,fixef=NULL,delta=1e-6,
                             invdelta=1/delta,
                             delta.nonlin=delta,
                             nugef=NULL,update.args=NULL,scale=NULL,
                             nonlin=NULL,parameters=NULL,
                             denseFixef=FALSE,
                             ignore.proj=FALSE){
  ## -------- Utility -----------
  "Dim<-" <- function(x,value){attr(x,"Dim") <- value;x}
  Dim <- function(x)attr(x,"Dim")
  is.covStruct <- function(x)"covStruct" %in% class(x)
  asList <- function(x){
    if(!is.list(x))structure(list(x),names=paste(names(x),collapse=":")) else x
  }

  ## * obs2cov:
  ## Factor attached to most covariance structures in the formula interface.
  ## Links observation entry with random-effect entry in the given covariance matrix.
  ## Its length is always equal to the number of observations and the number of levels is
  ## equal to the dimension of the covariance/precision matrix.
  ## * covMultiplier:
  ## Numeric of same length as "obs2cov".
  ## Multiply this (known) number on the random effect
  ## * More precisely: Let obs[i] be the ith obervation and eta the random effect of the given precision.
  ## Then log(E(obs[i])) gets the additive term multiplier[i]*eta[obs2cov[i]]

  obs2cov <- function(x)attr(x,"obs2cov")
  covMultiplier <- function(x)attr(obs2cov(x),"multiplier") ## (A copy of obs2cov then auto-copies covMultiplier!) 
  "obs2cov<-" <- function(x,value){
    attr(x,"obs2cov") <- value
    ## If no covMultiplier given then set it to one:
    lngt <- nrow(value)
    if(is.null(lngt))lngt <- length(value)
    if(is.null(covMultiplier(x)))covMultiplier(x) <- rep(1.0,lngt)
    x
  }
  "covMultiplier<-" <- function(x,value){
    lngt <- nrow(obs2cov(x))
    if(is.null(lngt))lngt <- length(obs2cov(x))
    if(lngt!=length(value))stop("Wrong multiplier length")
    attr(obs2cov(x),"multiplier") <- value
    x
  }
  
    
  ":" <- function(x,y){ ## Extend base ":" to accept NULL-arguments
    if(is.null(x) & is.factor(y))return(y)
    if(is.null(y) & is.factor(x))return(x)
    .Primitive(":")(x,y)
  }
  covterms <- list()
  saveCovstruct <- function(C,name)covterms[[name]] <<- C

  ## ========================== NOW COMES ALL THE METHODS:
  .methodEnv <- new.env()  ## This is where the overlodaded methods are located
  .methodEnv$PARM <- function(name="logscale"){
    expr <- substitute(exp(x),list(x=as.name(name)))
    p <- 0
    names(p) <- name
    C <- do.call(covParameter,list(expr,p))
    C
  }
  .methodEnv$ID <- function(fac,...,name){
    if(missing(name))name <- deparse(sys.call())
    C <- covIdent(n=nlevels(fac))
    ## Book-keeping (Make this general)
    Dim(C) <- nlevels(fac)
    names(Dim(C)) <- name
    obs2cov(C) <- fac
    C
  }
  ## Alternative: Add fixed effects in the ranef formula
  ## Only supports simple fixef formulaes given by a single factor
  ## (Following the line of the formula interface: every observation
  ##  must have a unique "lookup" in the ranef block).
  .methodEnv$FIXEF <- function(fac,...,name){
    ## Almost a copy of "ID"
    if(missing(name))name <- deparse(sys.call())
    C <- covIdent(x=rep(delta,nlevels(fac)))  ## <-- Flat prior
    Dim(C) <- nlevels(fac)
    names(Dim(C)) <- name
    obs2cov(C) <- fac
    C    
  }
  ## Call formula interface within formula interface on subset of data!
  .methodEnv$FORM <- function(...,
                              group=factor(1:nrow(data)),
                              name){
    if(missing(name))name <- deparse(sys.call())
    nd <- !duplicated(group)
    d <- data[nd,]
    fac <- factor(cumsum(nd)) ## key to lookup "data" in "d"
    C <- covStructFormula(...,data=d)
    dimC <- sum(unlist(lapply(Dim(C),prod)))
    levels(fac) <- seq(length=dimC)
    names(Dim(C)) <- paste(name,names(Dim(C)),sep=".")
    obs2cov(C) <- fac
    saveCovstruct(C,name)
    C
  }
  
  ## Similar to "ID" but with scale parameters
  .methodEnv$DIAG <- function(fac,...,name){
    if(missing(name))name <- deparse(sys.call())
    levelfac <- factor(paste(name,levels(fac),sep="."))
    C <- log(covDiag(levelfac))
    ## Book-keeping (Make this general)
    Dim(C) <- nlevels(fac)
    names(Dim(C)) <- name
    obs2cov(C) <- fac
    C
  }
  ## Apply general parameter transform
  ## -append name to new parameter names
  .methodEnv$transform <- function(C,expr=NULL,
                                   parameters=C@parameters,
                                   fix=NULL,
                                   range=NULL,
                                   name=NULL,
                                   ...){
    if(!is.null(range)){ ## Overwrite expr, parameters, etc
      if(!is.null(expr))warning("Ignoring argument expr")
      drange <- diff(range)
      mrange <- min(range)
      f <- function(x) drange/(1 + exp(-2 * x)) + mrange
      invf <- function(y) -0.5 * log(drange/(y - mrange) - 1)
      p <- invf(C@parameters)
      p[!is.finite(p)] <- 0
      names(p) <- paste("transf.", names(C@parameters), 
                        sep = "")
      expr0 <- substitute(
                          drange/(1 + exp(-2 * x)) + mrange
                          , list(drange=drange,mrange=mrange))
      fun <- function(name)do.call("substitute",list(expr0,list(x=as.name(name))))
      li <- lapply(names(p),fun)
      expr <- as.expression(li)
      names(expr) <- names(C@parameters)
      parameters <- p
    }
    ## Optionally append name
    if(is.character(name)){
      newnames <- paste(name,names(parameters),sep=".")
      subst <- lapply(newnames,as.name)
      names(subst) <- names(parameters)
      expr <- as.expression(lapply(expr,function(x)do.call("substitute",
          list(x, subst )) ))
      names(parameters) <- newnames
    }
    CC <- update(C,expr=expr,parameters=parameters,fix=fix,...)
    ## Copy attrributes (Dim,obs2cov,name,...)
    ## attributes(CC) <- attributes(C)
    Dim(CC) <- Dim(C)
    obs2cov(CC) <- obs2cov(C)
    ## Keep track of transform - or maybe not?
    attr(CC,"transform") <- list(C,expr=expr,parameters=parameters,fix=fix,...)
    ## Overwrite term
    name <- names(Dim(CC))
    saveCovstruct(CC,paste(name,collapse=":"))
    CC
  }
  .methodEnv$AR1 <- function(fac,...,name,reparam=TRUE){
    if(missing(name))name <- deparse(sys.call()[1:2])
    p <- .5
    names(p) <- name
    C <- covStructAR(p,n=nlevels(fac))
    if(reparam){
      f <- function(x)2/(1+exp(-2*x))-1
      invf <- function(y)-.5*log(2/(y+1)-1)
      p <- invf(C@parameters)
      names(p) <- paste("transf.",names(C@parameters),sep="")
      expr <- as.expression(substitute(2/(1+exp(-2*x))-1,
          list(x=as.name(names(p)))))
      names(expr) <- names(C@parameters)
      C <- update(C,expr,parameters=p)
    }
    ## Book-keeping (Make this general)
    Dim(C) <- nlevels(fac)
    names(Dim(C)) <- name
    obs2cov(C) <- fac
    saveCovstruct(C,name)
    C
  }
  .methodEnv$AR2 <- function(fac,...,name,reparam=TRUE){
    if(missing(name))name <- deparse(sys.call())
    p <- c(.05,-0.5)
    names(p) <- paste(name,c(1,2),sep="")
    C <- covStructAR(p,n=nlevels(fac))
    ## Natural param (phi1,phi2) lives on inner triangle {-1<phi2,phi2<1+phi1,phi2<1-phi1}.
    ## Mapped to:    (rho1,phi2) = (phi1/(1-phi2),phi2) with domain (-1,1)x(-1,1)
    if(reparam){
      p[1] <- p[1]/(1-p[2])
      names(p)[1] <- paste("corr.",names(C@parameters)[1],sep="")
      nm <- lapply(names(p),as.name)
      names(nm) <- c("rho1","phi2")
      expr <- as.expression(substitute(rho1*(1-phi2),nm))
      names(expr) <- names(C@parameters)[1]
      C <- update(C,expr,parameters=p[1])
    }
    
    ## Book-keeping (Make this general)
    Dim(C) <- nlevels(fac)
    names(Dim(C)) <- name
    obs2cov(C) <- fac
    saveCovstruct(C,name)
    C
  }

  .methodEnv$contAR2 <- function(t){
    name <- deparse(sys.call()[1:2])[1]
    tname <- deparse(sys.call()[[2]])[1]
    fac <- distFactor(t)
    dist <- abs(levelAttrib(fac)$dist[[1]])
    bm <- function(a,b,c,d)rbind(cbind(a,c),cbind(b,d))
    tmp <- lower.tri(dist)*2+upper.tri(dist)*3
    entry <- bm(dist*0+1,tmp,t(tmp),dist*0+4)
    dist.env <- list(t = bm(dist,dist,dist,dist),
                     I1 = entry==1,
                     I2 = entry==2,
                     I3 = entry==3,
                     I4 = entry==4)
    m <- as.matrix(abs(row(dist)-col(dist))<=1.5)
    P <- as(Matrix(bm(m,m,m,m)+0),"dgCMatrix")
    C <-
      covStruct(
                exp(a*t/s)*(cos(b*t/s)-a/b*sin(b*t/s))*I1+
                exp(a*t/s)*(-1/b*sin(b*t/s))*I2+
                exp(a*t/s)*(1/b*sin(b*t/s))*I3+
                exp(a*t/s)*(cos(b*t/s)+a/b*sin(b*t/s))*I4,
                dist=dist.env,pattern=P,
                parameters=c(a=-.99,b=sqrt(1-.99^2),s=1))
    Dim(C) <- c(nlevels(fac),2)
    names(Dim(C)) <- c(name,"deriv")
    levels(fac) <- seq(length=2*nlevels(fac))
    obs2cov(C) <- fac
    expr <- expression(
        a=-exp(shape)/(1+exp(shape)),
        b=sqrt(1-(exp(shape)/(1+exp(shape)))^2),
        s=exp(logscale))
    parameters=c(shape=0,logscale=0)
    C <- .methodEnv$transform(C,expr,parameters,name=name)
    C
  }
  
  ## pos: Factor mapping observations to spatial grid
  ## time: Factor mapping observations to time-grid
  .methodEnv$VAR1 <- function(pos,time,operator,...,name){
    if(missing(name))name <- deparse(sys.call())
    C <- covStructVAR1(operator,nlevels(time),...)
    p <- C@parameters
    ## Book-keeping (Make this general)
    Dim(C) <- c(nlevels(pos),nlevels(time))
    names(Dim(C)) <- c("pos","time")
    ## Note that "levels(x:y)" has "levels(y)" fastest running
    obs2cov(C) <- time:pos ## pos fastest running!
    saveCovstruct(C,name)
    C
  }
  .methodEnv$GMRF1 <- function(pos,pattern=attr(levelAttrib(pos)$grid,"pattern"),...,
                    name,log=TRUE){
    if(missing(name))name <- deparse(sys.call())
    p <- 1
    names(p) <- name
    if(!is(pattern,"sparseMatrix"))stop("Pattern must be a sparse matrix")
    C <- covStructGMRF(pattern,p)
    if(log)C <- log(C)
    Dim(C) <- nrow(pattern)
    names(Dim(C)) <- name
    obs2cov(C) <- pos
    saveCovstruct(C,name)
    C
  }

  ## Experimental: General unrestricted corr matrix
  .methodEnv$CORR <- function(fac,...,name,reparam=FALSE){
    if(missing(name))name <- deparse(sys.call()[1:2])
    C <- covStructGeneral(levels(fac),name,...)
    ## Book-keeping (Make this general)
    Dim(C) <- nlevels(fac)
    names(Dim(C)) <- name
    obs2cov(C) <- fac
    saveCovstruct(C,name)
    C
  }

  ## Scale covariance to correlation
  .methodEnv$COV2COR <- function(C){
    CC <- covStruct2Cor(C)
    C@fun <- CC@fun
    C@name <- CC@name
    C
  }
  
  ## Project covStruct to entries with data.
  ## Note: 1. projection may destroy sparseness.
  ## Note: 2. projection can be turned off with "ignore.proj=TRUE"
  .methodEnv$PROJ <- function(C,...,keep=NULL,name){
    if(missing(name))name <- deparse(sys.call(),500)[1] ##deparse(sys.call()[1:2])
    fac <- obs2cov(C)    
    newlev <- levels(factor(fac)) ## remove empty levels
    if(!is.null(keep)){
      ok <- all(keep %in% levels(fac))
      if(!ok)stop("PROJ: non-existing keep-levels")
      newlev <- c(keep,setdiff(newlev,keep)) ## "keep" goes up front
    }
    newfac <- factor(fac,levels=newlev)
    i <- match(levels(newfac),levels(fac)) ## entries to look up in old matrix
    CC <- covStructProj(C,i,...)
    Dim(CC) <- nlevels(newfac)
    names(Dim(CC)) <- name
    obs2cov(CC) <- newfac
    saveCovstruct(CC,name)
    CC
  }
  if(ignore.proj).methodEnv$PROJ <- function(C,...)C

  ## expr: expression or formula.
  .methodEnv$NONLIN <- function(expr,parameters,delta0=delta.nonlin,...,name){
    ## Variable names needed to calulate non-linear expression
    var.names <- setdiff(all.vars(expr),names(parameters))
    ## Get required variables and test they are *factors*
    dataEnvir <- parent.frame()
    vars <- lapply(var.names,get,dataEnvir)
    names(vars) <- var.names
    var.class <- sapply(vars,class)
    print(var.class)
    if(!all(sapply(vars,is.factor)))stop("Variables must be factors!")
    levs <- lapply(vars,levels)
    chr2num <- as.numeric
    numlevs <- lapply(levs,chr2num)
    cat("NOTE: Factor levels in non-linear effect converted to the following numerical values:\n")
    print(numlevs)
    e <- expand.grid(numlevs)
    fac <- interaction(vars,sep=":",lex.order=FALSE)
    funenv <- local(environment(),e)
    if(missing(name))name <- deparse(sys.call(),500)[1]
    f <- deriv(expr,names(parameters),function.arg=TRUE)
    environment(f) <- funenv
    C <- covStructNonlin(f,parameters,delta=delta0,...) ## dimension=nlevels(fac)+1
    ## Book-keeping
    Dim(C) <- c(asList(nlevels(fac)),list(1))
    names(Dim(C)) <- c(name,"FixThisParameterToOne")
    ##levels(fac) <- c(levels(fac),NA)
    levels(fac) <- c(levels(fac),"FixThisParameterToOne")
    obs2cov(C) <- fac
    saveCovstruct(C,name)
    C
  }
  ##local(debug(NONLIN),.methodEnv)  
  
  ## Arbitrary dense covStruct
  .methodEnv$COV <- function(fac,
                             formula=~exp(-exp(logb) * dist),
                             parameters=c(logb=0),
                             name,...){
    if(missing(name))name <- deparse(sys.call())[1]
    if(!is.factor(fac))stop("fac must be a factor")
    fac <- as(fac,"distFactor")
    pattern <- as(diag(nlevels(fac))*0+1,"dgCMatrix")
    cl <- call("covStruct",
               tail(formula,1)[[1]],
               parameters=parameters,
               pattern=pattern,
               dist=levelAttrib(fac)$dist)
    cl <- as.call(c(as.list(cl),list(...)))
    C <- eval(cl)
    dim <- attr(levelAttrib(fac)$dist,"Dim")
    Dim(C) <- dim
    names(Dim(C)) <- paste(name,names(dim),sep=".")
    obs2cov(C) <- fac
    saveCovstruct(C,name)
    C
  }

  .methodEnv$"*" <- function(x,y){
    if(is.numeric(x)&is.numeric(y))return(.Primitive("*")(x,y))
    C <- x%x%y
    Dim(C) <- c(Dim(y),Dim(x))
    ## Note that "levels(x:y)" has "levels(y)" fastest running
    obs2cov(C) <- obs2cov(x):obs2cov(y)
    covMultiplier(C) <-
      if      (is.null(covMultiplier(x)))covMultiplier(y)
      else if (is.null(covMultiplier(y)))covMultiplier(x)
      else covMultiplier(x)*covMultiplier(y)
    C
  }
  ## Scale entire covariance matrix - one variance parameter for each random effect!
  .methodEnv$SCALE <- function(C,fac=NULL,tag=NULL){
    if(is.null(fac)){
      e <- expand.grid(lapply(as.list(Dim(C)),seq))
      formals(paste)$sep <- ":"
      fac <- paste(paste(c(tag,names(e)),collapse=":"),do.call(paste,e),sep=".")
      fac <- factor(fac,levels=fac)
    }
    fac <- factor(fac)
    y <- C; x <- fac ## taken from ".methodEnv$:"
    C <- scaleCovStruct(y,x)
    Dim(C) <- Dim(y)
    obs2cov(C) <- obs2cov(y)
    return(C)
  }
  .methodEnv$":" <- function(x,y){
    if(is.factor(x) & is.factor(y))return(.Primitive(":")(x,y))
    if( is.factor(x)&is.covStruct(y) || is.factor(y)&is.covStruct(x)   ){
      if(!is.factor(x)){z <- x;x <- y;y <- z}
      C <- scaleCovStruct(y,x)
      Dim(C) <- Dim(y)
      obs2cov(C) <- obs2cov(y)
      return(C)
    }
    if( is.numeric(x)&is.covStruct(y) || is.numeric(y)&is.covStruct(x)   ){
      if(!is.numeric(x)){z <- x;x <- y;y <- z}
      C <- y
      covMultiplier(C) <- x
      return(C)
    }
    stop("Not yet implemented method for :")
  }
  .methodEnv$"+" <- function(x,y){
    C <- bdiagCovStruct(x,y)
    Dim(C) <- c(asList(Dim(x)),asList(Dim(y)))
    obs2cov(C) <- c(asList(obs2cov(x)),asList(obs2cov(y)))
    covMultiplier(C) <- c(asList(covMultiplier(x)),asList(covMultiplier(y)))
    C
  }
  
  ## =============================== Evaluation the formulas.
  ## Where to find variables: 1: data, 2: .methodEnv, 3: environment(formula)
  ## .methodEnv <- environment() ## This is where the overlodaded methods used to be located
  ## .thisEnv <- environment()
  .dataEnv <- local(environment(),data)
  myeval <- function(formula){
    .oldParentEnv <- parent.env(.methodEnv)
    on.exit(parent.env(.methodEnv) <- .oldParentEnv)
    parent.env(.methodEnv) <- environment(formula)
    parent.env(.dataEnv) <- .methodEnv
    eval(tail(formula,1)[[1]],.dataEnv)
  }
  
  ## Latent variables
  X <- myeval(formula)
  saveCovstruct(X,"ranefFormula")
  A <- Reduce(cbind2, Map(modMat,asList(obs2cov(X)), asList(covMultiplier(X))) )
  
  ## Add fixed effects:
  if(!is.null(fixef)){
    DimX <- Dim(X)
    ## Change - allow NAs in model.matrix. Replace NAs by zero.
    ## ---> A0 <- as(model.matrix(fixef,data),"dgCMatrix")
    mf <- model.frame(fixef,data=data,na.action=na.pass)
    mm <- model.matrix(fixef,mf)
    mm[is.na(mm)] <- 0
    A0 <- as(mm,"dgCMatrix")
    rm(mm,mf)  
    PRIOR <- covIdent(x=rep(delta,ncol(A0)),dense=denseFixef)
    X <- bdiagCovStruct(X,PRIOR)
    A <- cbind2(A,A0)
    Dim(X) <- c(asList(DimX),fixef=asList(ncol(A0)))
  } else Dim(X) <- c(asList(Dim(X)),fixef=0)

  ## Add observations:
  Y <- covIdent(x=rep(invdelta,nrow(data)))
  ## and nugget effect
  if(!is.null(nugef)){
      fac <- myeval(nugef)
      if(identical(fac,1))fac <- factor(rep("",nrow(data)))
      if(!is.factor(fac))stop("Nugget must be factor")
      levels(fac) <- paste("nugget",levels(fac),sep=".")
      Y <- log(covDiag(fac))
  }
  C <- axpyCovStruct(X,Y,A)
  
  if(!is.null(scale)){ ## Overwrite C
    fac <- myeval(scale)
    if(identical(fac,1))fac <- factor(rep("",nrow(data)))
    if(!is.factor(fac))stop("Scale must be factor")
    levels(fac) <- paste("scale",levels(fac),sep=".")
    C <- axpyCovStruct2(X,Y,A,Z=log(covDiag(fac))) 
  }
  
  ## --- Update covstruct?
  if(!is.null(update.args)){
    C <- do.call("update",c(C,update.args))
  }
  Dim(C) <- c(data=asList(nrow(data)),  asList(Dim(X)))

  ## --- Non linear extension?
  if(!is.null(nonlin)){
    oldDim <- Dim(C)
    f <- deriv(nonlin,names(parameters),function.arg=TRUE)
    environment(f) <- local(environment(),data)
    C <- covStructNonlinExtend(C,f,parameters,delta=delta.nonlin) ## dimension=nlevels(fac)+1
    Dim(C) <- c(oldDim,list(FixThisParameterToOne=1))
  }

  ## Further information
  if(!is.null(fixef)){
    attr(C,"fixefnames") <- colnames(A0)
  }
  ## When flat prior is assigned to the fixed effects one must
  ## add "-k*log(delta)" to logdet(full.precision).
  ## Thus "k/2*log(delta)" should be added to the negative log-likelihood
  k <- length(attr(C,"fixefnames"))
  attr(C,"negLogLikCorrection") <- k/2*log(delta)
  attr(C,"nfixed") <- k
  attr(C,"covterms") <- covterms  
  C
}


## Workaround: to get likelihood value
.lgc_getML <- function(n){ ## n=number of random effects
  ## Erase fixed effects from augmented precision:
  ## * insert 0 for index (i,j) if i>n or j>n.
  ## * However insert 1 if i==j
  eraseFE <- function(Q,n){
    row <- Q@i+1
    col <- rep(1:nrow(Q),diff(Q@p))
    ind <- (row>n)|(col>n)
    Q@x[ind] <- (row==col)[ind]
    Q
  }
  ## Restore control parameters on exit
  maxit.old <- .lgc_control()@maxit
  on.exit(.lgc_control(maxit=maxit.old))
  ## Set maxit=0
  .lgc_control(maxit=0L)
  Q <- .Call("lgc_getQ", PACKAGE="lgc") 
  Q <- eraseFE(Q,n)
  .Call("lgc_setQ", Q, PACKAGE="lgc")
  ok <- is.null(try(.Call("lgc_newton", PACKAGE="lgc"), silent = FALSE))
  ##.lgc_control() ## restore defaults
  if(!ok)return(NaN)
  return(as.numeric(.Call("lgc_getvalue", PACKAGE="lgc")))
}

##' Fit Log Gaussian (Cox) model using formula interface.
##' Can handle both count observations and/or direct oberservations of the Log Gaussian
##' process.
##' The latent field is integrated out using the Laplace approximation.
##' Fixed effects are estimated using a REML criterion.
##'
##' The formula interface helps building the often very large Gaussian Markov Random Fields.
##' These are represented by sparse matrices.
##'
##' @param ranef Passed to \code{covStructFormula}.
##' @param fixef Passed to \code{covStructFormula}.
##' @param data Passed to \code{covStructFormula}.
##' @param delta Passed to \code{covStructFormula}.
##' @param parscale Passed to \code{optim}.
##' @param control Object returned by \code{.lgc_control()}.
##' @param update.args Passed to \code{update} method of covStructs.
##' @param nugef Nugget effect formula.
##' @param etafix Formula giving a variable in the dataset with values in the random field to fix. Used to specify direct Log-Gaussian observations.
##'       NAs are ignored so that only parts of the field have to be observed.
##' @param permfun
##' @param scale
##' @param betafix Named vector of fixed effects to hold fixed.
##' @return Fitted object of class \code{fitlgc}.
fitlgc <- function(ranef,fixef,data,delta=1e-4,
                    parscale=rep(.01,length(C@parameters)),
                    control=.lgc_control(),update.args=NULL,
                    nugef=NULL,etafix=NULL,permfun=NULL,scale=NULL,
                    denseFixef=TRUE,betafix=NULL,
                    nonlin=NULL,parameters=NULL,
                    ignore.proj=FALSE,optimize=TRUE,...){
  dotargs <- list(...)
  covargs <- list(ranef,data=data,fixef,delta=delta,
                  update.args=update.args,nugef=nugef,scale=scale,denseFixef=denseFixef,
                  nonlin=nonlin,parameters=parameters,ignore.proj=ignore.proj)
  C <- do.call("covStructFormula",covargs)
  if(!is.null(permfun))dotargs$perm <- permfun(C) ## "permfun" could e.g. be "fastPerm" see below 
  
  Dim <- sapply(attr(C,"Dim"),prod)
  if(length(ranef)==2){ ## No counts - all missing
    cat("Note: No counts - all missing")
    N <- rep(NA,nrow(data))
  } else {
    if(length(ranef)!=3)stop("Random effects formula must have length 3")
    N <- eval(ranef[[2]],data)
  }
  N0 <- c(N,rep(NA,sum(Dim[-1])))
  d0 <- data.frame(N=N0)
  if(!is.null(etafix)){
    if(length(etafix)!=2)stop("Constrain random effects formula must have length 2")
    etafix <- eval(etafix[[2]],data)
    etafix <- c(etafix,rep(NA,sum(Dim[-1])))
  }
  ## Used to assume fixed effects where in the tail
  ## -not true anymore
  ## beta.index <- tail(1:sum(Dim),attr(C,"nfixed"))
  beta.index <- as.vector(.lgc_splitIndex(Dim=attr(C,"Dim"))$fixef)
  
  if(!is.null(betafix)){
    nm <- attr(C,"fixefnames")
    if(!is.vector(betafix) | is.null(names(betafix)) | !all(names(betafix)%in%nm)){
      cat("\'betafix\' must be a named vector. Allowed names are\n")
      print(nm)
      stop()
    }
    if(is.null(etafix))etafix <- rep(NA,sum(Dim)) ## Now etafix exists
    etafix[beta.index] <- betafix[nm]
  }
  ## Non-linear effects - fix multipliers to one:
  csDim <- cumsum(Dim)
  FixThisParameterToOneIndex <- csDim[names(csDim)=="FixThisParameterToOne"]
  if(length(FixThisParameterToOneIndex)>0){
    if(is.null(etafix))etafix <- rep(NA,sum(Dim)) ## Now etafix exists
    etafix[FixThisParameterToOneIndex] <- 1
  }
  ## Fit model
  eval(control)

##   fit <- fitlgc(N~1,pos=~lon+lat,data=d0,
##                 covstruct=C,lgcsparse=TRUE,parscale=parscale,etafix=etafix,...)
  args <- list(N~1,pos=~lon+lat,data=d0,
               covstruct=C,lgcsparse=TRUE,parscale=parscale,etafix=etafix,
               optimize=optimize)
  args <- c(args,dotargs)
  ##fit <- do.call("fitlgc",args)
  fit <- do.call(".lgc_optimize",args)
  
  ## We want "data" - not "d0":
  fit$data <- data
  
  fit$negLogLikCorrection <- attr(C,"negLogLikCorrection") ## Add to value
  fit$nfixed <- attr(C,"nfixed")
  fit$Dim <- attr(C,"Dim")
  fit$fixefnames <- attr(C,"fixefnames")
  if(!is.null(etafix))fit$etafix <- na2sparse(etafix)

  if(optimize){
    ## Full Hessian
    if(!denseFixef)fullHessian2 <- fullHessian ## Grab old method
    fit <- c(fit, try(fullHessian2(fit, beta.index), silent = !TRUE))
  }
    
  ## Workaround to get likelihood
  nranef <- sum(head(sapply(fit$Dim,prod),-1))
  fit$valueML <- .lgc_getML(nranef)

  ## Used by predictFull
  fit$call <- match.call()
  
  ## Get fill-reduciong permutation (may want this later)
  fit$perm <- .Call("lgc_getperm", PACKAGE="lgc")
  
  class(fit) <- "fitlgc"
  fit
}
logLik.fitlgc <- function(object,...)-(object$valueML)

.lgc_splitIndex <- function(object=NULL, x=seq(length=sum(sapply(Dim,prod))), Dim=object$Dim ){
  len <- sapply(Dim,prod)
  fac <- factor(x=rep.int(1:length(len),len),levels=1:length(len),labels=names(len))
  li <- split(x,fac)
  ans <- Map(array,li,Dim)
  if(!is.null(object)){
    names(ans$fixef) <- object$fixefnames
  }
  ans  
}

fitted.fitlgc <- function(object,...){
  .lgc_splitIndex(object,x=object$eta)
}
coef.fitlgc <- function (object, ...){
  c(fitted(object)$fixef,object$par)
}
vcov.fitlgc <- function (object, ...){
  solve(object$hessian.full)
}


## Allow user to implement permutation methods.
## A permutation method takes a covstruct-object as imput
## and returns a permutation.
fastPerm <- function(C){
  n <- sapply(attr(C,"Dim"),prod)
  te <- attr(C,"covterms")
  Q0 <- evaluate(te$ranefFormula)
  .lgc_symbolic( Q0 )
  perm <- .Call("lgc_getperm", PACKAGE="lgc")
  ntail <- sum(n)-(n[1]+nrow(Q0))
  p <- c( 0:(n[1]-1), ## not permute data
          n[1]+perm,  ## permute ranef
          seq(from=sum(n)-ntail,length=ntail) ) ## not permute fixef
  as.integer(p)
}

## simulate covStruct returned by "covStructFormula".
## Output result as a list of arrays like "fitted.fitlgc" would do.
setMethod("simulate","covStruct",function(object, nsim = 1, seed = NULL, ...){
  if(nsim!=1)stop("Currently only nsim=1 implemented")
  Q <- as(evaluate(object),"dsCMatrix")
  x <- GMRFsample(Q,...)
  obj <- list(Dim=attr(C,"Dim"),eta=x)
  fv <- fitted.fitlgc(obj)
  fv
})

##' Draw simulations from a fitted object of class 'fitlgc'.
##'
##' This function can simulate at any level of the model. If \code{type}=='data' then
##' count data will be simulated with parameters as in the fitted object.
##' If \code{type}=='prior' then the prior random field will be simulated, including any term
##' of the model formula along with the random effects directly associated with a data-entry.
##' By default, the simulation will be convieniently split in a list with these components
##' (\code{as.list}==TRUE).
##' Finally, if \code{type}=='posterior' then a posterior sample (i.e. the random field given the data)
##' will be returned.
##' By default, the fixed effects of the model are deterministic (without variation) in
##' the case of \code{type}=='data' and \code{type}=='prior', and non-deterministic in the case of
##' 'posterior' sampling.
##' @title LGC simulation.
##' @param object Object of class 'fitlgc'
##' @param nsim Number of simulations.
##' @param seed Simulation seed.
##' @param type What to simulate: 'data', 'prior' or 'posterior'
##' @param term What model term to simulate ('prior' simulation only)
##' @param betafix Should REML estimates be fixed at their estimated values?
##' @param as.list Return simulation in convenient list-format?
##' @param ...
##' @return The simulation
simulate.fitlgc <- function (object,nsim = 1, seed = NULL,
                             type=c("data", "prior", "posterior"),
                             term=NULL,
                             fixbeta=(type!="posterior"),
                             as.list=(type!="data")&(nsim==1)&(is.null(term)),
                             ##super=TRUE,
                             ##etafix=object$etafix,
                             ...){
  fit <- object
  type <- match.arg(type)
  if(!is.null(seed))set.seed(seed)
  if(is.null(term)){
    if(nsim>1 & as.list)stop("Can not return list if more than one sample.")
    ## Update etafix
    etafix <- fit$etafix
    if(fixbeta){
      Dim <- sapply(attr(fit$covstruct,"Dim"),prod)
      if (is.null(etafix)) etafix <- na2sparse(rep(NA, sum(Dim)))
      beta.index <- as.vector(.lgc_splitIndex(fit)$fixef)
      etafix[beta.index] <- fit$eta[beta.index]
    }
    fit$etafix <- etafix
    ## Update data
    obsind <- which(!is.na(fit$N))
    if(type=="prior"|type=="data")fit$N[] <- NA
    ## Draw posterior sample (= prior if no data)
    .lgc_reconstruct(fit)
    etahat <- as.vector(.Call("lgc_geteta", PACKAGE="lgc"))
    n <- length(etahat)
    u <- matrix(rnorm(n*nsim), n, nsim)
    etasim <- as.matrix(.Call("lgc_GMRFsampleH", u, PACKAGE="lgc")) + etahat
    if(type=="data"){
      ans <- exp(etasim[obsind,])
      ans[] <- rpois(length(ans),ans)
    } else {
      ans <- etasim
    }
    if(as.list)ans <- .lgc_splitIndex(fit,ans)
    return(ans)
  } else {
    if(type!="prior")warning("Individual term simulation only allowed for prior model - type argument changed.")
    if(as.list)stop("Individual term simulation can not return list.")
    covterms <- attr(object$covstruct,"covterms")
    if(!(term %in% names(covterms))){
      cat("Allowed term names:\n")
      print(names(covterms))
      stop("Invalid term name")
    }
    Q <- evaluate(covterms[[term]],object$par)
    Q <- as(Q,"dsCMatrix")
    eta <- GMRFsample(Q,n=nsim,super=TRUE)
    return(eta)
  }
}

simulate.fitlgc.old <- function (object, nsim = 1, seed = NULL,
                              term="ranefFormula",super=TRUE,
                              posterior=FALSE,etafix=object$etafix,...){
  if(posterior){ ## Posterior sampling from *full* model
    C <- object$covstruct
    Q <- evaluate(C,object$par)
    dataentry <- !is.na(as.vector(object$N))
    diag(Q) <- diag(Q)+exp(object$eta)*dataentry
    Q <- as(Q,"dsCMatrix")
    .lgc_symbolic(Q,object$perm)
    if(!is.null(etafix)).Call("lgc_setetafix", etafix, PACKAGE="lgc")
    .Call("lgc_factorize", PACKAGE="lgc")
    u <- matrix(rnorm(nrow(Q)*nsim),nrow(Q))
    eta <- as.matrix(.Call("lgc_GMRFsample",u, PACKAGE="lgc"))+object$eta
    return(eta)
  }
  covterms <- attr(object$covstruct,"covterms")
  Q <- evaluate(covterms[[term]],object$par)
  Q <- as(Q,"dsCMatrix")
  eta <- GMRFsample(Q,n=nsim,super=super)
  eta
}

## plot parameter estimates with conf int
plot.fitlgc <- function(fit,which=grep(pattern,names(coef(fit))),
                         pattern=NULL,cex.axis=.6,...){
  ## Choose subset
  par <- coef(fit)
  sd <- sqrt(diag(vcov(fit)))
  if(missing(which) & is.null(pattern))which <- 1:length(par)
  par <- par[which]
  sd <- sd[which]
  nm <- names(par)
  
  ## Plot
  plot(par,axes=FALSE,xlab="",...)
  axis(1,at=1:length(par),labels=nm,cex.axis=cex.axis,las=3)
  axis(2,cex.axis=cex.axis,las=1)
  ci <- 1.96*sd
  for(i in 1:length(par))segments(i,par[i]-ci[i],i,par[i]+ci[i])
  box()
}

## Utility:
fullHessian <- function(obj,beta.index,mis=beta.index){

  ## Make sure that report=TRUE and REML=FALSE
  ## (and restore original settings on exit)
  ## -----------------------------------------
  experimental <- getOption("lgc_control")@experimental
  report <- getOption("lgc_control")@report
  on.exit(.lgc_control(experimental=experimental,report=report))
  .lgc_control(report=1L)
  bits <- intToBits( experimental )
  if(bits[2]){
    bits[2] <- as.raw(0) ## Turn off 2nd bit
    .lgc_control(experimental=packBits(bits,"integer"))
  }

  ## Make sure that Q, Qdot, etahat, H, etc. correspond
  ## to the parmeters of fitted object "obj"
  ## --------------------------------------------------------
  covstruct <- obj$covstruct
  thetav <- obj$par
  covenv <- do.call(covstruct@fun, as.list(thetav))
  ok <- is(try(covenv$precision), "sparseMatrix")
  stopifnot(ok)
  .Call("lgc_setQ", covenv$precision, PACKAGE="lgc")
  ok <- is.null(try(.Call("lgc_newton", PACKAGE="lgc"), silent = FALSE))
  .Call("lgc_gradient", covenv$Dprecision, PACKAGE="lgc")
  
  ## Now we may assume everything is fine
  ## ------------------------------------
  D2f.eta.eta <- .Call("lgc_getH", PACKAGE="lgc")
  expr1 <- GMRFmarginal(D2f.eta.eta,mis,perm=TRUE,super=TRUE)
  ## Hessian of profile (marginal precision of theta)
  Hprof <- as(obj$hessian,"dsCMatrix")
  ## Linear relation between theta and beta
  Gbeta.theta <- as( .Call("lgc_getgradeta", PACKAGE="lgc")[mis,,drop=FALSE] ,"dgCMatrix")
  ## Full Hessian
  H <- GMRFextend(expr1,Gbeta.theta,Hprof)
  H <- as(H,"matrix")
  par <- c(obj$eta[mis],obj$par)

  beta <- obj$eta[mis]
  Hbeta <- as(GMRFmarginal(as(H,"dsCMatrix"),1:length(beta)),"matrix")
  list(par.full=par, hessian.full=H, beta=beta, hessian.beta=Hbeta)
}


fullHessian2 <- function(obj,beta.index,mis=beta.index){

  ## Make sure that report=TRUE and REML=FALSE
  ## (and restore original settings on exit)
  ## -----------------------------------------
  experimental <- getOption("lgc_control")@experimental
  report <- getOption("lgc_control")@report
  on.exit(.lgc_control(experimental=experimental,report=report))
  .lgc_control(report=1L)
  bits <- intToBits( experimental )
  if(bits[2]){
    bits[2] <- as.raw(0) ## Turn off 2nd bit
    .lgc_control(experimental=packBits(bits,"integer"))
  }

  ## Make sure that Q, Qdot, etahat, H, etc. correspond
  ## to the parmeters of fitted object "obj"
  ## --------------------------------------------------------
  .lgc_derivmode(TRUE)
  covstruct <- obj$covstruct
  thetav <- obj$par
  covenv <- do.call(covstruct@fun, as.list(thetav))
  ok <- is(try(covenv$precision), "sparseMatrix")
  stopifnot(ok)
  Q <- covenv$precision  
  .Call("lgc_setQ", Q, PACKAGE="lgc")
  ok <- is.null(try(.Call("lgc_newton", PACKAGE="lgc"), silent = FALSE))
  .Call("lgc_gradient", covenv$Dprecision, PACKAGE="lgc")
  
  ## Now we may assume everything is fine
  ## ------------------------------------
  ## D2f.eta.eta <- .Call("lgc_getH")
  ## expr1 <- GMRFmarginal(D2f.eta.eta,mis,perm=TRUE,super=TRUE)
  if(length(beta.index)>0){
    T <- .Call("lgc_invH", PACKAGE="lgc")[beta.index,beta.index] ## permuted triangular
    S <- T+t(T)
    diag(S) <- .5*diag(S)
    expr1 <- solve(S)
  }
  else expr1 <- Matrix(0,0,0)
  expr1 <- as(expr1,"dgCMatrix")
  ## Hessian of profile (marginal precision of theta)
  Hprof <- as(obj$hessian,"dsCMatrix")
  ## Linear relation between theta and beta
  Gbeta.theta <- as( .Call("lgc_getgradeta", PACKAGE="lgc")[mis,,drop=FALSE] ,"dgCMatrix")
  ## Full Hessian
  H <- GMRFextend(expr1,Gbeta.theta,Hprof)
  H <- as(forceSymmetric(H),"matrix")
  par <- c(obj$eta[mis],obj$par)

  beta <- obj$eta[mis]
  Hbeta <- as(GMRFmarginal(as(H,"dsCMatrix"),1:length(beta)),"matrix")
  list(par.full=par, hessian.full=H, beta=beta, hessian.beta=Hbeta)
}

summary.fitlgc <- function(object,...){
  df <- data.frame(estimate=coef(object),sd=sqrt(diag(vcov(object))))
  df$"P(Chi^2>|x|)" <- pchisq((df$estimate/df$sd)^2,df=1,lower.tail=FALSE)
  df
}

## Reconstruct internal stuff of a fitted object
## To be used in:
## - simulate.fitlgc
## - aggregate intensities
## - fullHessian
.lgc_reconstruct <- function(fit){
  fit$covstruct@parameters <- fit$par
  tmp <- .lgc_sparse(N=fit$N,covstruct=fit$covstruct,perm=fit$perm,etafix=fit$etafix,etastart=fit$eta)
  .Call("lgc_newton", PACKAGE="lgc")
}

## extract internal parameter transform of fitted object.
## Useful with "simconfint" e.g.:
## simconfint(fit,.lgc_parTransform(fit))
.lgc_parTransform <- function(fit){
  te <- attributes(fit$covstruct)$covterms
  li <- lapply(te,function(x)attributes(x)$transform$expr)
  names(li) <- NULL
  myexpr <- do.call("c",li)
  function(par)sapply(myexpr,eval,as.list(par))
}

## Update fit by ignoring PROJ statements in formula interface
fullPredict <- function(fit){
  cl <- fit$call
  cl$data <- fit$data
  cl$ignore.proj <- TRUE
  cl$optimize <- FALSE
  cl$th0 <- fit$par
  fit2 <- eval(cl)
  dnm <- setdiff(names(fit),names(fit2))
  fit2[dnm] <- fit[dnm]
  fit2
}
