##' Log-gaussian cox processes
##'
##' Fit log-gaussian cox process by maximum likelihood using the
##' Laplace approximation. Formula interface for model construction incudes
##' basic autoregressive, vector-autogressive and spatial autoregressive
##' building blocks, along with arbitrary composition in terms of addition and
##' separable extensions. Mainly targeted aquatic applications.
##' 
##' \tabular{ll}{ Package: \tab lgc\cr Type: \tab Package\cr Version: \tab
##' 1.4\cr Date: \tab 2008-01-09\cr License: \tab GPL\cr
##' LazyLoad: \tab yes\cr } 
##' 
##' @name lgc-package
##' @aliases lgc-package lgc
##' @docType package
##' @author Kasper Kristensen
##' 
##' Maintainer: <kaskr@@imm.dtu.dk>
##' @examples
##' \dontrun{
##' demo(NorthSeaCod)
##' }
NULL

.lgc_optimize <- function(formula,pos,data,subset,
                           covstruct,
                           hessian=TRUE,
                           use.gradient=TRUE,
                           optimize=TRUE,
                           check=c("none","derivatives","likelihood"),
                           th0, ## Optional: initial parameter
                           recordfile=TRUE, ## Record "par" and "value" ?
                           parscale=rep(1,length(th0)),
                           optim.maxit=300,
                           optim.control=list(maxit=optim.maxit,trace=3,reltol=1e-12,REPORT=1,
                             parscale=rep(parscale,length.out=length(th0))),
                           perm,             ## User permutation for lgc_sparse
                           etafix=NULL,
                           ...){

  ## Initialize
  check <- match.arg(check)
  N <- as.numeric(data$N)
  lgc <- .lgc_sparse(N,covstruct,perm,etafix=etafix)
  theta <- covstruct@parameters
  
  ## ---------------------------------------------------------------
  ## Recording mode ?
  ## ---------------------------------------------------------------
  tracefile <- recordfile
  record <- function(x,file){
    if(!tracefile)return(NULL)
    if(is.list(x)|is.environment(x))return(NULL)
    x <- as.vector(x)
    write(x,file,append=TRUE,ncolumns=length(x),sep=";")
  }
  if(tracefile){
    tmpfile <- tempfile(pattern="fitlgc")
    likfile <- paste(tmpfile,"lik",sep=".")
    gradfile <- paste(tmpfile,"grad",sep=".")
    curfile <- paste(tmpfile,"cur",sep=".")
    cat("Recording to files:\n", likfile, "\n", gradfile, "\n")
  }

  ## ---------------------------------------------------------------
  ## Store likelihood value at first evaluation - fn(th0)
  ## ---------------------------------------------------------------
  value.init <- NULL
  storeIfFirstValue <- function(value){
    if(is.null(value.init))value.init <<- value
  }
  
  ## ---------------------------------------------------------------
  ## Likelihood function and gradient.
  ## ---------------------------------------------------------------  
  fn <- function(th,report=FALSE){
    record(th,curfile)
    ans <- lgc(theta=th,
               gradient=FALSE,
               report=report)
    record(c(th,ans),likfile)
    storeIfFirstValue(ans)
    ans
  }
  gr <- function(th){
    res <- lgc(theta=th,
               gradient=TRUE,
               report=FALSE)
    ans <- res
    record(c(th,max(abs(ans))),gradfile)
    ans
  }

  ## ---------------------------------------------------------------
  ## Initial parameter
  ## ---------------------------------------------------------------
  th1 <- theta
  if(!missing(th0)){
    nm <- intersect(names(th1),names(th0))
    th1[nm] <- th0[nm]
  }
  th0 <- th1
  

  ## ------- If recording write parameter names to files
  names.theta <- names(th0)
  record(c(names.theta,"value"),likfile)
  record(c(names.theta,"mgc"),gradfile)
  record(c(names.theta),curfile)
  ## ---------------------------------------------------
  
  if(check=="derivatives"){
    require(numDeriv)
    return(list(exact=gr(th0),fd=grad(fn,th0)))
  } else
  if(check=="likelihood"){
    return(list(th0=th0,fn=fn,gr=gr))
  }
  if(!use.gradient)gr <- NULL
  
  if(optimize){
    opt <- optim(th0,fn,gr,method="BFGS",hessian=hessian,
                 control=optim.control
                 )
  } else opt <- list(par=th0,value=fn(th0))
  opt <- c(opt,fn(opt$par,report=TRUE))
  opt$value.init <- value.init
  opt$N <- N
  opt$covstruct <- covstruct  
  opt$data <- data
  if(hessian){
    opt$vcov <- try(solve(opt$hessian),silent=TRUE)
    if(is.matrix(opt$vcov))opt$sd <- sqrt(diag(opt$vcov))
  }
  class(opt) <- "fitlgc"
  opt
}


print.fitlgc <- function (x,na.print="",corr.digits=2,...) 
{
    cat("Object of class 'fitlgc'\n")
    corr <- try(round(suppressWarnings(cov2cor(x$vcov)),corr.digits),silent=TRUE)
    dimnames(corr) <- NULL
    corr[upper.tri(corr,diag=TRUE)] <- NA
    if(identical(class(corr),"try-error"))corr <- NA
    m <- cbind(x$par,x$sd,corr)
    colnam <- rep("",ncol(m))
    colnam[1:3] <- c("par","sd","corr")
    colnames(m) <- colnam
    print(m,na.print=na.print,...)
    invisible(m)
}


