## --------------------------------------------------------------------------  
## Current Methods:
##   GMRFextend
##   kronecker
## Consider also
##   "+" - sparse matrix addition (because pattern generation is
##         time consuming compared to flops)
##   "%*%" - probably not so much gained
##   "bdiag" - not much gained I think because allready fast (pattern trivial)
.lgc_opt <- local({
  ## trace on/off
  trace <- function(on=TRUE){
    if(on){
      .lgc_opt$print <- NULL
      .lgc_opt$cat <- NULL
    } else
    {
      .lgc_opt$print <- function(...)NULL
      .lgc_opt$cat <- function(...)NULL
    }
  }
  ## Restore environment to its state before calling "activateOptMethods"
  cleanup <- function(env=.lgc_opt){
    fun <- function(ch){
      obj <- get(ch,env)
      if(is.environment(obj)){cleanup(obj);rm(list=ch,envir=env)}
      else if(is.function(obj)&&identical(environment(obj),.lgc_opt)) NULL
      else rm(list=ch,envir=env)
    }
    nm <- ls(env,all=TRUE)
    lapply(nm,fun)
  }
  ## -------------------------------------
  ## Turn on openmp (linux gcc) ?
  ## -------------------------------------
  openmp <- function(on=TRUE){
    if(on)Sys.setenv(PKG_CPPFLAGS="-fopenmp",PKG_LIBS="-lgomp")
    else Sys.unsetenv(c("PKG_CPPFLAGS","PKG_LIBS"))
  }

  ## ============== Overview of allocated workspace and methods
  methodSummary <- function(f,expr=
                            data.frame(Gb=as.vector(object.size(answerList)/1e9),
                                       allocated=length(free),
                                       free=sum(free))){
    expr <- substitute(expr)
    fun <- function(f,expr){
      e <- environment(f)
      if(is.null(e$signature))return(NULL)
      c(list(eval(expr,e)),fun(e$nextmethod,expr))
    }
    ans <- fun(f,expr)
    if(is.null(ans))return(NULL)
    do.call("rbind",ans)
  }
  summary <- function(...){
    FUN <- function(x)methodSummary(x,...)
    rbind(
          do.call("rbind",eapply(.lgc_opt,FUN)),
          do.call("rbind",eapply((.lgc_opt$methodEnv),FUN))
          )
  }

  
  init <- function(){
    require(inline)
    ## -------------------------------------
    ## Fast version of "x[i]<-x[i]+y"
    ## -------------------------------------
    code <- "
       int n=LENGTH(i);
       if(n!=LENGTH(y))error(\"ups\");
       int *pi=INTEGER(i);
       double *px=REAL(x);
       px--;
       double *py=REAL(y);
   #pragma omp parallel for schedule(static)
       for(int j=0;j<n;j++)px[pi[j]]+=py[j];
       return R_NilValue;
     "
    subaddassign <- cfunction(signature(x="numeric", i="integer", y="numeric"), code)
    ## -------------------------------------
    ## Fast version of "x[i]<-y"
    ## -------------------------------------
    code <- gsub("\\+=","=",code)
    subassign <- cfunction(signature(x="numeric", i="integer", y="numeric"), code)
    ## -------------------------------------
    ## Fast version of "x[]<-y[i]*z[j]"
    ## Used for kronecker product where e.g.
    ## i is found by replacing the numerical
    ## values Y@x by seq(Y@x) and replacing
    ## Z@x by ones and calulating (Y%x%Z)@x !
    ## Likewise for j ...
    ## -------------------------------------
    code <- "
       int n=LENGTH(i);
       if((n!=LENGTH(j)) | (n!=LENGTH(x)))error(\"ups\");
       int *pi=INTEGER(i);
       int *pj=INTEGER(j);
       double *px=REAL(x);
       double *py=REAL(y);
       double *pz=REAL(z);
       py--;pz--;
   #pragma omp parallel for schedule(static)
       for(int k=0;k<n;k++)px[k]=py[pi[k]]*pz[pj[k]];
       return R_NilValue;
     "
    assignsubmult <- cfunction(signature(x="numeric",
                                         y="numeric",i="integer",
                                         z="numeric",j="integer" ),code)
    ## -------------------------------------
    ## Fast version of "x[]<-0"
    ## -------------------------------------
    code <- "
       int n=LENGTH(x);
       double *px=REAL(x);
   #pragma omp parallel for schedule(static)
       for(int j=0;j<n;j++)px[j]=0.0;
       return R_NilValue;
     "
    erase <- cfunction(signature(x="numeric"), code)
    ## -------------------------------------
    ## Object duplication
    ## Equivalent of (and not much faster than)
    ## duplicate <- function(x)new(class(x),x)
    ## -------------------------------------
    code <- "
       SEXP res;
       res = Rf_duplicate(x);
       return res;
     "
    duplicate <- cfunction(signature(x="any"), code)
    ## -------------------------------------
    ## Weak S4 object duplication
    ## Creates a new object and copies
    ## slot-*references*.
    ## -------------------------------------
    weakCopy <- local({
      code <- "
       copyMostAttrib(x,y);
       return y;
     "
      copy <- cfunction(signature(x="any",y="any"), code)
      function(x){
        copy(x,new(class(x)))
      }
    })
    ## Set attrib "env" by reference
    code <- "
       setAttrib(x,install(\"env\"),y);
       return R_NilValue;
     "
    setEnvAttrib <- cfunction(signature(x="any",y="any"), code)
    ## Experimental: New memory allocation strategy:
    ## Let S4-Matrix objects share all slots except x-slot.
    ## Set xslot to NULL
    code <- "
       setAttrib(x,install(\"x\"),R_NilValue);
       return x;
     "
    rmxslot <- cfunction(signature(x="any"), code)
    ## Input: S4 object ("dgCMatrix") without x-slot
    ## Output: Same object with new allocated x-slot.
    S4alloc <- local({
      code <- "
       copyMostAttrib(x,y);
       int n=LENGTH(getAttrib(x,install(\"i\")));
       setAttrib(y,install(\"x\"),allocVector(REALSXP,n));
       return y;
     "
      copy <- cfunction(signature(x="any",y="any"), code)
      function(x){
        copy(x,new(class(x)))
      }
    })

    ## -------------------------------------
    ## Primitive method lookup/construct
    ## -------------------------------------
    ## expr *must* construct a function "makeTemplate"
    ## And *must* return a function.
    ## User method *must* apply the functions
    ## "registerReturnValue" and "allocateReturnValue".
    generateNewMethod <- function(expr,isExpression=FALSE){
      if(!isExpression)expr <- substitute(expr)
      signature <- NULL
      delayedAssign("nextmethod",generateNewMethod(expr,TRUE))
      getPatternSignature <- function(...){
        info <- function(x){ ## Currently used info to collect - may be extended.
          if(is.list(x))return(do.call("rbind",lapply(x,info)))
          .length <- function(x)if(is(x,"sparseMatrix"))length(x@x) else length(x)
          .row <- function(x){
            if(is(x,"sparseMatrix"))
              c(first=x@i[1],last=x@i[max(length(x@i),1)])
            else
              c(first=1,last=nrow(x))
          }
          c(class=class(x),dim=dim(x),length=.length(x),row=.row(x))
        }
        ## do.call("rbind",lapply(list(...),info))
        ## Replacing old code with "list(...)" to avoid copying 
        e <- userfunGetvars(...)
        li <- lapply(userfunVarnames,function(x)info(e[[x]]))
        do.call("rbind",li)
      }
      userenv <- new.env() ## All userfun stuff goes here
      userfun <- eval(expr,userenv) 
      makeTemplate <- userenv$makeTemplate
      if(is.null(makeTemplate))stop("Missing makeTemplate")
      ## Below is fix of issue with list(...) copying arguments
      userfunVarnames <- setdiff(names(formals(userfun)),"...")
      userfunGetvars <- args(userfun)
      body(userfunGetvars) <- quote(environment())
      ## Manage expandable list of allocated return values
      answerList <- list()
      ## NEW: Only need one - share pointers.
      registerReturnValue <- function(ans){
        ans <- rmxslot(ans)
        answerList[[1]] <<- ans
      }
      allocateReturnValue <- function(){
        S4alloc(answerList[[1]])
      }      
      free <- logical()
      retfun <- function(...){
        sig <- getPatternSignature(...)
        id <- identical(sig,signature)
        if(id){
          cat("Succes: Method found for signature\n")
          print(sig)
          cat("userfun start computing..\n")
          ans <- userfun(...)
          cat("userfun done\n")
          return(ans)
        }
        else if(is.null(signature)){
          cat("No match: Creating template of signature\n")
          print(sig)
          makeTemplate(...)
          signature <<- sig
          return(userfun(...))
        }
        else {
          cat("No method found here - trying next method \n")
          return(nextmethod(...))
        }
      }
      retfun
    }
    
    ## ============== USER METHODS
    .lgc_opt$bdiag <- generateNewMethod({
      INDEX <- NULL
      xslot <- function(x)if(is.matrix(x))x else x@x
      makeTemplate <- function(li){
        b <- mybdiag(li)
        b@x[] <- seq(b@x)
        le <- sapply(li,function(x)length(xslot(x)))
        fac <- factor(rep(seq(le),le),levels=seq(le))
        INDEX <<- lapply(split(b@x,fac),as.integer)
        registerReturnValue(b)
      }
      function(li){
        ans <- allocateReturnValue()
        for(i in seq(li))subassign(xslot(ans),INDEX[[i]],xslot(li[[i]]))
        ans
      }
    })
    
    .lgc_opt$kronecker <- generateNewMethod({
      i <- NULL
      j <- NULL
      xslot <- function(x)if(is.matrix(x))x else x@x
      makeTemplate <- function(X,Y,...){
        if(is.matrix(X))X <- as(X,"dgCMatrix")
        if(is.matrix(Y))Y <- as(Y,"dgCMatrix")
        X@x[] <- 1
        Y@x[] <- seq(Y@x)
        ans <- X%x%Y
        j <<- as.integer(ans@x)
        X@x[] <- seq(X@x)
        Y@x[] <- 1
        ans <- X%x%Y
        i <<- as.integer(ans@x)
        erase(ans@x)
        registerReturnValue(ans)
      }
      function(X,Y,...){
        ans <- allocateReturnValue()
        assignsubmult(ans@x,xslot(X),i,xslot(Y),j)
        ans
      }
    })

    .lgc_opt$GMRFextend <- generateNewMethod({
      INDEX <- NULL
      makeTemplate <- function (Q, A, R){
        V <- -Q %*% A
        AtQA <- -crossprod(A, V)
        V@x[] <- 1
        Vt <- t(V)
        Vt@x[] <- 2
        AtQA@x[] <- 4
        if (!missing(R)){
          R@x[] <- 8
          AtQA <- AtQA + R
        }
        Q@x[] <- 16
        ans <- lgc::block2x2(Q, Vt, V, AtQA)
        ind <- list(V=1,Vt=2,Q=16,AtQA=c(4,12),R=c(8,12))
        myf <- function(x)as.integer(which(ans@x%in%x))
        INDEX <<- lapply(ind,myf)
        registerReturnValue(ans)
      }
      function (Q, A, R){
        if (!missing(R)){R <- as(R,"dgCMatrix")}
        ans <- allocateReturnValue()
        erase(ans@x)
        V <- -Q %*% A
        AtQA <- -crossprod(A, V)
        subaddassign(ans@x,INDEX$Q,Q@x)
        subaddassign(ans@x,INDEX$V,V@x)
        subaddassign(ans@x,INDEX$Vt,t(V)@x)
        subaddassign(ans@x,INDEX$AtQA,AtQA@x)
        if (!missing(R))subaddassign(ans@x,INDEX$R,R@x)
        ans
      }
    })

    .lgc_opt$block2x2 <- generateNewMethod({
      INDEX <- NULL
      makeTemplate <- function (A,B,C,D){
        A@x[] <- 1
        B@x[] <- 2
        C@x[] <- 3
        D@x[] <- 4
        ans <- cbind2(rbind2(A, B), rbind2(C, D))
        ind <- list(A=1,B=2,C=3,D=4)
        myf <- function(x)as.integer(which(ans@x==x))
        INDEX <<- lapply(ind,myf)
        registerReturnValue(ans)
      }
      function (A,B,C,D){
        ans <- allocateReturnValue()
        ##erase(ans@x)
        subassign(ans@x,INDEX$A,A@x)
        subassign(ans@x,INDEX$B,B@x)
        subassign(ans@x,INDEX$C,C@x)
        subassign(ans@x,INDEX$D,D@x)
        ans
      }
    })
    
  }
  environment()
})

##' Turn on optimized sparse matrix construction (Only relevant for very large models).
##'
##' Speed up model fitting by caching sparse matrix patterns and sharing pointers between them.
##' Currently requires package \code{inline}.
##'
##' Building large sparse or semisparse precision matrices can be
##' very time consuming for large models - even though major flops takes place elsewhere - after
##' the matrices have been built.
##' During model fitting, matrices of identical sparseness pattern are computed repeatedly with
##' different numerical values. By caching the required matrix operations we can preallocate an arsenal of
##' sparse matrix templates and thereby eliminate this overhead. Furthermore, different matrices with
##' identical sparseness pattern can share most sparse matrix slots - thereby reducing memory.
##' Before model fitting call \code{activateOptMethods()} to compile and overload the methods. 
##' Use \code{.lgc_opt$trace(FALSE)} to turn off tracing information.
##' After model fitting use
##' \code{.lgc_opt$summary()} to get an overview of currently cached methods, and
##' \code{.lgc_opt$cleanup()} to cleanup workspace - Imortant before e.g. saving
##' a fitted object!
##' 
##' @param on 
##' @return NULL
##' @examples
##' \dontrun{
##' ##' === Separable correlation on 1000x1000 grid
##' ##' Without optimization
##' C <- covStructAR(n=1000);CC <- C%x%C
##' system.time(Q <- evaluate(CC))
##' ##' Now with optimization
##' activateOptMethods() ## Turn on
##' C <- covStructAR(n=1000);CC <- C%x%C
##' system.time(Q <- evaluate(CC)) ## First call generates template
##' system.time(Q <- evaluate(CC)) ## Following calls are fast
##' .lgc_opt$summary() ## List of active methods
##' activateOptMethods(FALSE) ## Turn off
##' }
activateOptMethods <- function(on=TRUE){
  if(!on){
    .lgc_opt$cleanup()
    return(NULL)
  }
  .lgc_opt$openmp()
  .lgc_opt$init()
  ## Alternative
  local({
    methodEnv <- new.env()
    methodEnv$"%x%" <- function (X, Y) kronecker(X, Y)
    methodEnv$mybdiag <- .lgc_opt$bdiag
  },.lgc_opt)
  ## Changing environment of S4 method...
  fun <- "kronecker"
  sig <- rep("covStruct",2)
  m <- selectMethod(fun,sig)
  m <- m@.Data
  environment(m) <- .lgc_opt$methodEnv
  setMethod(fun,sig,m,where=environment())
  ## Switch env for ordinary function
  NAMESPACE <- asNamespace("lgc")
  sapply( ls( envir = NAMESPACE ), unlockBinding, env = NAMESPACE )
  environment(axpyCovStruct) <<- .lgc_opt$methodEnv
  environment(bdiagCovStruct) <<- .lgc_opt$methodEnv
  environment(axpyCovStruct2) <<- .lgc_opt$methodEnv
  environment(addNugget) <<- .lgc_opt$methodEnv
  environment(covStructNonlinExtend) <<- .lgc_opt$methodEnv
  ## combining lists by reference
  .lgc_opt$methodEnv$c.list <- local({
    code <- "
       int i;
       int nx=LENGTH(x);
       for(i=0;i<LENGTH(x);i++)SET_VECTOR_ELT(ans,i,VECTOR_ELT(x,i));
       for(i=0;i<LENGTH(y);i++)SET_VECTOR_ELT(ans,i+nx,VECTOR_ELT(y,i));
       return ans;
     "
    comb <- cfunction(signature(x="any",y="any",ans="any"), code)
    c2 <- function(x,y){
      ans <- vector("list",length(x)+length(y))
      names(ans) <- c(names(x),names(y))
      comb(x,y,ans)
    }
    function(x,y,...){
      if(missing(y))x else c2(x,c.list(y,...))
    }
  })
  ## List subassign by reference
  .lgc_opt$methodEnv$"[[<-" <- local({
    code <- "
       SET_VECTOR_ELT(x,INTEGER(i)[0]-1,y);
       return R_NilValue;
     "
    subassign <- cfunction(signature(x="any",i="integer",y="any"), code)
    function(x,i,value){
      subassign(x,as.integer(i),value)
      x
    }
  })

}


