## Implementing termwise goodness of fit test:

## The smart way:
## ==============
## S ~ nugget precision
## R ~ block matrix including beta parameters.
## Full precision Q:
## [S        -SA  ]
## [-AtS  R+AtSA  ]
## has cholesky L=
## [LS       ]
## [-AtLS  LR]
## and thus if r=(r1,r2) then
## Lt*r = ( LSt(r1-A*r2)) , LRt*r2 )
## Note that LR is a block matrix with Cholesky of each term.
##
## 1. Start by calling .lgc_reconstruct. We then have internaly
##    * LQ (without penalty if some parms are fixed). 
##    * LH (including penalty)
##    * etahat which accounts for etafix such that if some of the betas
##      are fixed then this prepagates correctly to the data entries.
## 2. Draw sample from N(0,H^-1). Add etahat. This is a posterior sample
##    "r=etasim" subject to etafix. 
##    LQt * r contains conponents:
##      LSt * [I  -A]*r = Direct goodness of fit test for nugget effect.
##      LRt * r2 = Goodness of fit for each term independently.
##                 (final component - fixef - should just be ignored)
## ==========================================
## Seems we have a bug in simulate posterior when setting etafix.
## Internally etafix only change the precision of H. Not Q.
## That is: LQ is unaffected. LH is changed so that we have close to
##  zero variations on etafix entries.
## When sampling lgc_GMRFsample then LQ is used thus etafix is not fixed!
## ==========================================
## Mayby better to use .lgc_reconstruct ?
## And then lgc_GMRFsampleH ?

##' Simulation based residuals for goodness of fit.
##'
##' Generate an approximate posterior sample standardized with the prior.
##'
##' A goodness of fit test for the log Gaussian Cox model can be performed using
##' the following simple principle: Draw \emph{one} sample \eqn{\eta^*} from the
##' posterior distribution \eqn{\eta|x}. Standardize with the prior \eqn{u=\Sigma^{-1/2}\eta^*}.
##' If the prior is correctly specified then \eqn{u} is a vector of independent standard normal
##' variates - which can be further investigated by a Kolmogorov-Smirnov test or a Chi-square test.
##' The simulation seed is deliberately fixed to stress the assumption of \emph{one sample}.
##' @param model A fitted model object.
##' @param fixbeta Logical. Should the fixed effects be deterministic in the sample?
##' @param seed Simulation seed.
##' @return A list with a vector of standardized residuals for each model term.
rstandard.fitlgc <- function(model,fixbeta=TRUE,seed=123){
  fit <- model
  if(!is.null(seed))set.seed(seed)
  ## 1. Change etafix
  etafix <- fit$etafix
  if(fixbeta){
    Dim <- sapply(attr(fit$covstruct,"Dim"),prod)
    if (is.null(etafix)) etafix <- na2sparse(rep(NA, sum(Dim)))
    beta.index <- as.vector(.lgc_splitIndex(fit)$fixef)
    etafix[beta.index] <- fit$eta[beta.index]
  }
  fit$etafix <- etafix
  ## 2. Draw posterior sample
  .lgc_reconstruct(fit)
  etahat <- as.vector(.Call("lgc_geteta"))
  n <- length(etahat)
  u <- matrix(rnorm(n), n)
  etasim <- as.matrix(.Call("lgc_GMRFsampleH", u)) + etahat
  ## 3. Standardize with prior
  LQ <- .Call("lgc_getLQ")
  LQm <- as(LQ,"sparseMatrix")
  p <- LQ@perm+1
  ip <- p;ip[p] <- seq(along=p)
  L <- LQm[ip,p]
  u <- t(L)%*%etasim
  ## 4. Return in convenient format
  .lgc_splitIndex(fit,u)
}
