#include "cs_utils.h"
# ifdef _OPENMP
#include <omp.h>
# endif


SEXP CHMfactor_to_sparse(SEXP x);

SEXP get_factor_pattern(SEXP obj, char *pat, int offset)
{
    SEXP facs = GET_SLOT(obj, Matrix_factorSym), nms;
    int i;

    if (!LENGTH(facs)) /* Fast return for empty factor slot: */
	return R_NilValue;
    nms = getAttrib(facs, R_NamesSymbol);
    for (i = 0; i < LENGTH(nms); i++) {
	const char *nm = CHAR(STRING_ELT(nms, i));
	if (strlen(nm) > offset && !strcmp(pat + offset, nm + offset))
	    return VECTOR_ELT(facs, i);
    }
    return R_NilValue;
}


/* Copy a sparse matrix */
cs* copy(cs* A){
  cs* ans=cs_spalloc (A->m,A->n,A->nzmax,1, /*triplet*/ 0);
  int i;
  for(i=0;i<=A->n;i++)ans->p[i]=A->p[i];
  for(i=0;i<A->nzmax;i++){
    ans->i[i]=A->i[i];
    ans->x[i]=A->x[i];
  }
  return ans;
}


/* Kronecker product */
/* a11*B  a12*B  */
/* a21*B  a22*B  */
/* Gives segfault - why? -> moved to ops.c  */
SEXP kronecker_old(SEXP x, SEXP y){
  cs* A=AS_CSP(x);
  cs* B=AS_CSP(y);
  cs* C=cs_spalloc(A->m*B->m,A->n*B->n,A->nzmax*B->nzmax,1, /*triplet*/ 0);

  int i,j,p,q,k=0,l=0;
  double *Ax=A->x,*Bx=B->x,*Cx=C->x;
  int *Ai=A->i,*Bi=B->i,*Ci=C->i,*Ap=A->p,*Bp=B->p;
  /* The (i,j)-column of C contains "outer(B[,j],A[,i])" */
  for(i=0;i<A->n;i++) /* loop over columns (i,j) of C */
    for(j=0;j<B->n;j++){
      C->p[l]=k;      /* k=current non-zero position */
      l++;            /* l=current column of C */
      for(p=Ap[i];p<Ap[i+1];p++)
	for(q=Bp[j];q<Bp[j+1];q++){
	  Cx[k]=Ax[p]*Bx[q];
	  Ci[k]=Ai[p]*B->m+Bi[q];
	  k++;
	}
    }
  C->p[l]=k; /* Finalize p */

  //Free(A); Free(B);
  return Matrix_cs_to_SEXP(C,"dgCMatrix", /* cs_spfree(C) */ 1);
}  


/* -------------- OpenMP version of kronecker --------------- */

/* Compute A(:,i) o B and place result in i'th mxn-block of C */
void kronecker_column(cs* A, cs* B, cs* C, int i){
  double *Ax=A->x,*Bx=B->x,*Cx=C->x;
  int *Ai=A->i,*Bi=B->i,*Ci=C->i,*Ap=A->p,*Bp=B->p;
  int k=B->nzmax*(Ap[i]-Ap[0]);
  int l=i*B->n;
  for(int j=0;j<B->n;j++){
    C->p[l]=k;      /* k=current non-zero position */
    l++;            /* l=current column of C */
    for(int p=Ap[i];p<Ap[i+1];p++)
      for(int q=Bp[j];q<Bp[j+1];q++){
	Cx[k]=Ax[p]*Bx[q];
	Ci[k]=Ai[p]*B->m+Bi[q];
	k++;
      }
  }
}
SEXP kronecker_parallel(SEXP x, SEXP y){
  cs* A=AS_CSP(x);
  cs* B=AS_CSP(y);
  cs* C=cs_spalloc(A->m*B->m,A->n*B->n,A->nzmax*B->nzmax,1, /*triplet*/ 0);
#pragma omp parallel for
  for(int i=0;i<A->n;i++){ 
    kronecker_column(A,B,C,i);
  }
  C->p[C->n]=A->nzmax*B->nzmax; /* Finalize p */
  return Matrix_cs_to_SEXP(C,"dgCMatrix", /* cs_spfree(C) */ 1);
}  


/* Inner product of sparse matrix-columns */
/* < A(:,i) , B(:,j) >  */
/* Assumes *sorted* row indices  */
/* (can be obtained by double transposing the matrices) */
double inner_prod(cs *A, cs *B, int i, int j){
  int *Ap=A->p, *Bp=B->p, *Ai=A->i, *Bi=B->i, pa, pb;
  double *Ax=A->x, *Bx=B->x, sum=0.0;
  pb=Bp[j];
  for(pa=Ap[i];pa<Ap[i+1];pa++){
    while( Bi[pb]<Ai[pa] & pb<Bp[j+1] )pb++;
    if( Bi[pb]==Ai[pa] & pb<Bp[j+1] )sum+=Ax[pa]*Bx[pb];
  }
  return sum;
}


/* <dense,sparse> inner product */
/* < A(:) , B(:,j) > for rowindex>=i */
/* Just the inner_pruduct if i==0 */
double inner_prod2(double *A, cs *B, int i, int j){
  int *Bp=B->p, *Bi=B->i, p;
  double *Bx=B->x, sum=0.0;
  /* Note: Only Bi[p]>i needed  */
  for(p=Bp[j];(p<Bp[j+1])&(Bi[p]<i);p++);
  for(;p<Bp[j+1];p++){
    sum+=Bx[p]*A[Bi[p]];
  }
  return sum;
}


/* <dense,sparse> inner product  < A(:) , B(:,j) > */
/* (so that length of A must be B->m) */
/* j=column of B to consider */
/* pstart=pointer to nonzero element of B we want to start with */
/* (if pstart=B->p[j] its just an ordinary inner product)  */
double inner_prod3(double *A, cs *B, int j, int p_start){
  int *Bp=B->p, *Bi=B->i, p;
  double *Bx=B->x, sum=0.0;
  for(p=p_start;p<Bp[j+1];p++){
    //printf("p_start: %d p: %d col: %d row: %d\n",p_start,p,j,Bi[p]);
    sum+=Bx[p]*A[Bi[p]];
  }
  return sum;
}


/* Find pointer to transpose entry */
int pointer_transpose(const cs *A, int *pt){
  int *w,*w2,p,m,n,*Ap,*Ai,j;
  m=A->m; n=A->n; Ap=A->p; Ai=A->i;
  w=cs_calloc(A->m,sizeof(int));  /* initialized to zero */
  w2=cs_calloc(A->m+1,sizeof(int));  /* dummy workspace */
  for(p=0;p<Ap[n];p++)w[Ai[p]]++; /* row counts */
  cs_cumsum(w2,w,m); /* row pointers */
  for(j=0;j<n;j++){
    for(p=Ap[j];p<Ap[j+1];p++){
      pt[p]=w[Ai[p]]++;
    }
  }
  cs_free(w); cs_free(w2);
  return 0;
}


/* Diagonal of sparse matrix-product diag(AB). */
/* Assumes *sorted* row indices */
SEXP diag_matprod(SEXP x, SEXP y){
  cs* A=AS_CSP(x);
  cs* B=AS_CSP(y);
  cs* AT = cs_transpose (A, 1) ;
  SEXP ans;
  PROTECT(ans=allocVector(REALSXP,A->n));
  for(int i=0;i<A->n;i++)REAL(ans)[i]=inner_prod(AT,B,i,i);
  cs_spfree(AT); //Free(A); Free(B);
  UNPROTECT(1);
  return ans;
}
  

/* keep off-diagonal entries; drop diagonal entries */
static int cs_diag (int i, int j, double aij, void *other) { return (i != j) ; }
static int cs_upper (int i, int j, double aij, void *other) { return (i <= j) ; }

/* dgCMatrix -> dsCMatrix */
SEXP dgC2dsC(SEXP x){
  cs* A=AS_CSP(x);
  cs* Acp=copy(A);
  cs_fkeep (Acp, &cs_upper, NULL) ;
  return Matrix_cs_to_SEXP(Acp,"dsCMatrix",1);
}

// NOT USED
/* Compute cholesky  */
SEXP chol(SEXP x, SEXP ord){
  cs* A=AS_CSP(x);
  int order=INTEGER(ord)[0];
  
  css *S ;
  csn *N ;
  int n, ok ;
  if (!CS_CSC (A)) return (0) ;	   /* check inputs */
  n = A->n ;
  S = cs_schol (order, A) ;	   /* ordering and symbolic analysis */
  N = cs_chol (A, S) ;             /* numeric Cholesky factorization */
  ok = (S && N) ;
  if(!ok){cs_sfree(S);cs_nfree(N);/*Free(A);*/error("chol failure");}
  cs* L=copy(N->L);
  cs_sfree (S) ;
  cs_nfree (N) ;
  return Matrix_cs_to_SEXP(L,"dgCMatrix",1);
}

/* Interface to cs_cholsol */
SEXP cholsol(SEXP x, SEXP b, SEXP ord){
  cs* A=AS_CSP(x);
  int order=INTEGER(ord)[0];
  SEXP ans;
  PROTECT(ans=duplicate(b));
  int ok=cs_cholsol(order,A,REAL(ans));
  //Free(A);
  UNPROTECT(1);
  if(!ok)error("cholsol unable to solve system");
  return ans;
}


/* Often we want to simulate from N(0,Sigma) only knowing the  */
/* sparse precision  */
/* So given u~N(0,I) and precision Q. */
/* Write PQP^t = LL^t . Then X=P^t L^t^-1 u has variance Sigma */
/* To do that we simply have to erase two lines of cs_cholsol!!!  */
int cs_cholsol2 (const cs *A, double *b)
{
  int order=1;
  double *x ;
  css *S ;
  csn *N ;
  int n, ok ;
  if (!CS_CSC (A) || !b) return (0) ;	    /* check inputs */
  n = A->n ;
  S = cs_schol (order, A) ;		    /* ordering and symbolic analysis */
  N = cs_chol (A, S) ;		    /* numeric Cholesky factorization */
  x = cs_malloc (n, sizeof (double)) ;    /* get workspace */
  ok = (S && N && x) ;
  if (ok)
    {
      //cs_ipvec (S->pinv, b, x, n) ;	/* x = P*b */
      //cs_lsolve (N->L, x) ;		/* x = L\x */
      for(int i=0;i<n;i++)x[i]=b[i];    /* x=b */
      cs_ltsolve (N->L, x) ;		/* x = L'\x */
      cs_pvec (S->pinv, x, b, n) ;	/* b = P'*x */
    }
  cs_free (x) ;
  cs_sfree (S) ;
  cs_nfree (N) ;
  return (ok) ;
}
/* Interface to cs_cholsol2 */
SEXP cholsol2(SEXP x, SEXP b){
  cs* A=AS_CSP(x);
  SEXP ans;
  PROTECT(ans=duplicate(b));
  int ok=cs_cholsol2(A,REAL(ans));
  //Free(A);
  UNPROTECT(1);
  if(!ok)error("cholsol unable to solve system");
  return ans;
}



/* ---------------------- inv_new ------------------------*/
/* A: Matrix of class "dsCMatrix" with cached Cholesky. */
/* A: Matrix of class "sparseMatrix". */
/* I have made it possible to supply an extra argument "SigmaStart" */
/* containing a pre-copmputed lower-right corner of Sigma in the permuted */
/* domain. The subset-of-inverse function then computes the rest of Sigma */
/* starting from the top of the lower-right specified corner. */

SEXP inv(SEXP A, SEXP perm, SEXP SigmaStart){
  /* Convert the cached Cholesky object to the corresponding lower */
  /* triangular matrix in Csparse format. Also get the permutation. */

  // Change: Assume the CHMfactor is input - not dsCMatrix with factors
  //SEXP C=get_factor_pattern(A,"...Cholesky",3);
  //SEXP C=A;
  int *P, *Pinv, n;
  //cs* L=AS_CSP(CHMfactor_to_sparse(C));
  cs* L=AS_CSP(A);
  //P = INTEGER(GET_SLOT(C, Matrix_permSym));
  P = INTEGER(perm);
  n = L->n;
  Pinv = cs_pinv(P,n);

  /* Compute sparsity pattern of (subset of) Sigma: (L+LT) */
  cs* LT = cs_transpose (L, 1) ;	/* compute L' */
  cs *Sigma,*tmp;
  tmp=cs_add(L,LT,0.0,0.0);
  Sigma=cs_transpose(tmp,1); /* Transpose symmetric matrix to sort columns */
  cs_spfree(tmp);cs_spfree(LT);

  /* If SigmaStart is specified */
  int start=n-1; /* Default is to start on last row in outer loop */
  if(SigmaStart != R_NilValue){
    /* TODO: Check that pattern of SigmaStart is a subset of Sigma-pattern */
    /* before adding those */
    tmp=AS_CSP(SigmaStart);
    for(int i=0;i<tmp->nzmax;i++)start=CS_MIN(start,tmp->i[i]);
    start--;
    cs *tmp2=Sigma;
    Sigma=cs_add(tmp,tmp2,1.0,1.0);
    cs_spfree(tmp2);
    tmp2=Sigma;
    Sigma=cs_transpose(tmp2,1); /* Transpose to sort columns */
    cs_spfree(tmp2);
  }

  /* Workspace to hold *dense* column of L0 */
  double *L0 = cs_calloc(n, sizeof(double));

  /* Permutation to map p'th element of Sigma to transposed entry  */
  int *ptrans = cs_calloc(Sigma->nzmax, sizeof(int));
  pointer_transpose(Sigma,ptrans);
  
  /* Initialize */
  int i,j,k,p,pt,nL0;
  int *Li=L->i, *Lp=L->p, *Sigmai=Sigma->i, *Sigmap=Sigma->p; 
  double *Lx=L->x, *Sigmax=Sigma->x;
  double sigij,Lii,sum;
  for(p=0;p<n;p++)L0[p]=0.0; /* clear L0 column - not needed (calloc) */

  /* Recusive algorithm to find subset of inverse matrix */
  /* Note: Only loop over upper triangle of Sigma (j>=i) */
  /* The inner loop may be done in parallel for all n>=j>i. */
  /* Only for j=i are the previous values within the inner loop needed. */
  /* To compute Sigma(i,j) we need the inner product <L0(:,i),Sigma(:,j)> */
  /* i.e. we need to know Sigma only strictly below the (i,j)-entry */
  /* The result is most naturally stored in Sigma(j,i) - ensuring that  */
  /* when we reach j=i all Sigmas are known strictly below the (i,i)-entry */
  /* So that Sigma(i,i) can be computed. */
  /* When we have all the Sigma(j,i)'s (j>=i) they must be copied to the */
  /* transposed entries Sigma(i,j). */
  /* We choose to store the entire symmetric matrix - and not just the lower
     triangle - in order to have good memory access in the inner loop */
  for(i=start;i>=0;i--){ // Changed "n-1" to "start" !!!
    Lii=Lx[Lp[i]];

    /* scatter L(:,i) over L0 */
    for(p=Lp[i]+1;p<Lp[i+1];p++)L0[Li[p]]=Lx[p];
    
    /* Number of nonzeros strictly below element Sigma(i,i) */
    nL0=Lp[i+1]-Lp[i]-1;

    /* openmp loop j=n,...,i+1 */
#pragma omp parallel for private(sigij,j,pt) 
    for(p=Sigmap[i+1]-1; p>Sigmap[i+1]-1 -nL0 ; p--){
      j=Sigmai[p];   /* column index */
      pt=ptrans[p];  /* Pointer to transpose entry*/
      //sigij=-inner_prod2(L0,Sigma,i,j)/Lii;
      sigij=-inner_prod3(L0,Sigma,j,pt)/Lii;
      
      Sigmax[p]=sigij;
      Sigmax[pt]=sigij;
    }
    /* Now i==j */
    p=Sigmap[i+1]-1 -nL0;
    j=Sigmai[p]; /* ==i */
    //sigij=-inner_prod(L0,Sigma,i,j)/Lii;
    //sigij=-inner_prod2(L0,Sigma,i,j)/Lii;
    sigij=-inner_prod3(L0,Sigma,j,p)/Lii;
    sigij+=1.0/(Lii*Lii);
    Sigmax[p]=sigij;
    
    /* Clear L0 */
    for(p=Lp[i]+1;p<Lp[i+1];p++)L0[Li[p]]=0.0;
  }
  
  /* Permute Sigma back - and transpose to sort columns! */
  cs* pSigma=cs_permute(Sigma,P,Pinv,1);
  cs_spfree(Sigma);
  Sigma=cs_transpose(pSigma,1);
  cs_spfree(pSigma);
  cs_free(Pinv);
  cs_free(ptrans);
  cs_free(L0);
  return Matrix_cs_to_SEXP(Sigma,"dgCMatrix",1);
}


/* ----------- NEW VERSION FILLING "UP" INSTEAD OF "LEFT" ----- */
/* ======> NOT PARALLEL - BUT MORE CACHE FRIENDLY ? */
/* WARNING: Does not support a "SigmaStart"   */

/* Recursion: Compute Sigma(:,k) assuming known entries below diagonal */
/* Sigma(i,k) = 1_(k==i)/(Lii*Lii) - 1/Lii*sum(L0(:,i)*Sigma(:,k))  */
void recursion(cs *L, cs *Sigma, int k, int *ptrans, int delay, int* col_done){
  int i,n=Sigma->n;
  int *Sp=Sigma->p, *Lp=L->p, *Si=Sigma->i, *Li=L->i, ps, pdiag, ptop=Sp[k];
  double *Sx=Sigma->x, *Lx=L->x, Lii;

  /* Workspace to hold *dense* Sigma column. *Must* be cleared */
  double *S = cs_calloc(n, sizeof(double));
  for(i=0;i<n;i++)S[i]=0.0;

  /* If "delay" is non-zero we have to wait until col k+1 is done.
     Note: we need an extra flush to make sure we have the updated col_done 
     *and* the updated values of Sigma */

  if(delay){
    //printf("col %d waiting\n",k);
    while(col_done[0]!=(k+1)){
#pragma omp flush(col_done)
    }
#pragma omp flush(col_done,Sx)
    //printf("col %d done waiting\n",k);
  }

  /* Scatter Sigma column - and find diagonal pointer */
  for(ps=Sp[k];ps<Sp[k+1];ps++){
    S[Si[ps]]=Sx[ps];
    if(Si[ps]==k)pdiag=ps;
  }

  /* If "delay" is non-zero we must prepare data for next thread
     and make that data visible */
  if(delay){  
    for(int p=pdiag;p>=ptop;p--){
      i=Si[p];
      if(i<k-delay)break;
      Lii=Lx[Lp[i]];
      S[i]=-1/Lii*inner_prod3(S,L,i,Lp[i]+1);
      if(k==i)S[i]+=1.0/(Lii*Lii);
      Sx[ptrans[p]]=S[i];
    }
    col_done[0]=k;
#pragma omp flush(col_done,Sx)
  }

  /* Compute dense column */
  for(int p=pdiag;p>=ptop;p--){
    i=Si[p];
    Lii=Lx[Lp[i]];
    S[i]=-1/Lii*inner_prod3(S,L,i,Lp[i]+1);
    if(k==i)S[i]+=1.0/(Lii*Lii);
  }

  /* Write to sparse and transposed entries */
  for(int p=pdiag;p>=ptop;p--){
    i=Si[p];
    Sx[p]=S[i];
    Sx[ptrans[p]]=S[i];
  }

  /* Now Sx is updated and future threads must see the changes */
  if(delay){
#pragma omp flush(Sx)
    //printf("completed col %d\n",k);
  }

  cs_free(S);
}



SEXP inv_up(SEXP A, SEXP perm, SEXP SigmaStart){
  /* Convert the cached Cholesky object to the corresponding lower */
  /* triangular matrix in Csparse format. Also get the permutation. */

  // Change: Assume the CHMfactor is input - not dsCMatrix with factors
  //SEXP C=get_factor_pattern(A,"...Cholesky",3);
  //SEXP C=A;
  int *P, *Pinv, n;
  //cs* L=AS_CSP(CHMfactor_to_sparse(C));
  cs* L=AS_CSP(A);
  //P = INTEGER(GET_SLOT(C, Matrix_permSym));
  P = INTEGER(perm);
  n = L->n;
  Pinv = cs_pinv(P,n);

  /* Compute sparsity pattern of (subset of) Sigma: (L+LT) */
  cs* LT = cs_transpose (L, 1) ;	/* compute L' */
  cs *Sigma,*tmp;
  tmp=cs_add(L,LT,0.0,0.0);
  Sigma=cs_transpose(tmp,1); /* Transpose symmetric matrix to sort columns */
  cs_spfree(tmp);cs_spfree(LT);

  /* If SigmaStart is specified */
  int start=n-1; /* Default is to start on last row in outer loop */
  if(SigmaStart != R_NilValue){
    /* TODO: Check that pattern of SigmaStart is a subset of Sigma-pattern */
    /* before adding those */
    tmp=AS_CSP(SigmaStart);
    for(int i=0;i<tmp->nzmax;i++)start=CS_MIN(start,tmp->i[i]);
    start--;
    cs *tmp2=Sigma;
    Sigma=cs_add(tmp,tmp2,1.0,1.0);
    cs_spfree(tmp2);
    tmp2=Sigma;
    Sigma=cs_transpose(tmp2,1); /* Transpose to sort columns */
    cs_spfree(tmp2);
  }

  /* Workspace to hold *dense* column of L0 */
  double *L0 = cs_calloc(n, sizeof(double));

  /* Permutation to map p'th element of Sigma to transposed entry  */
  int *ptrans = cs_calloc(Sigma->nzmax, sizeof(int));
  pointer_transpose(Sigma,ptrans);
  
  /* Initialize */
  int i,j,k,p,pt,nL0;
  int *Li=L->i, *Lp=L->p, *Sigmai=Sigma->i, *Sigmap=Sigma->p; 
  double *Lx=L->x, *Sigmax=Sigma->x;
  double sigij,Lii,sum;
  for(p=0;p<n;p++)L0[p]=0.0; /* clear L0 column - not needed (calloc) */


  /* Parallel column version  */
  int col_done;
  int delay=0;
# ifdef _OPENMP
  //delay=omp_get_num_threads();
  delay=omp_get_max_threads();
# endif
  //printf("delay set to %d\n",delay);

  if(delay){ /* Parallel version */
    /* Columns above a known lower-right corner can be computed with delay=0 */
#pragma omp parallel for schedule(static,1)
    for(i=n-1;i>start;i--){
      recursion(L,Sigma,i,ptrans,0,&col_done);
    }
    /* Now do the rest of the columns */
    col_done=start+1;
#pragma omp barrier
#pragma omp parallel for schedule(static,1)  
    for(i=start;i>=0;i--){ // Changed "n-1" to "start" !!!
      recursion(L,Sigma,i,ptrans,delay,&col_done);
    }
  } else { /* Serial version */
    for(i=n-1;i>=0;i--){
      recursion(L,Sigma,i,ptrans,0,&col_done);
    }
  }

  
  /* Permute Sigma back - and transpose to sort columns! */
  cs* pSigma=cs_permute(Sigma,P,Pinv,1);
  cs_spfree(Sigma);
  Sigma=cs_transpose(pSigma,1);
  cs_spfree(pSigma);
  cs_free(Pinv);
  cs_free(ptrans);
  cs_free(L0);
  return Matrix_cs_to_SEXP(Sigma,"dgCMatrix",1);
}



