
/*
  DESCRIPTION:
  This code performs the Laplace approximation for the (discretized) 
  log guassian cox process, defined as:

  * eta = (eta_1,...,eta_n) is a mean-zero latent gaussian vector with 
  inverse variance matrix Prec(eta)=Q.
  
  * N = (N_1,...,N_n) is a vector of counts associated with eta such that
  the conditional distribution of N_k|eta is Poisson(exp(eta_k)). Possibly
  N can have a fixed pattern of missing values.

  * Q is an n by n sparse positive definite matrix.

  The full likelihood function is (up to a normalizing constant)

  L(Q|N,eta) ~ |Q|^.5 exp(-1/2 eta' Q eta) * prod_i=1^n exp(eta_i)^N_i exp(-exp(eta_i))

  The desired marginal likelihood L(Q|N) is defined as the integral of L(Q|N,eta) 
  wrt. eta.

  FUNCTIONS:
  "lgc_newton": Finds the optimum "etahat" and the Laplace approximation of 
  the marginal likelihood function.
  "lgc_gradient": Finds the derivative of the marginal likelihood function.

  OPTIONAL:
  Soft constraints can be imposed on the random field eta. A soft constraint
  modifies the quadratic form 
  ".5 eta' Q eta" to 
  ".5 eta' Q eta + .5 (eta-etafix)' P (eta-etafix)"
  where P is chosen such that
  * P[i,i] is huge whenever we want eta[i]=etafix[i].
  * P[i,i] is zero otherwise
  so that the modified quadratic form effectively describes the conditional distribution 
  of eta given certain fixed entries.
  Rearranging terms we get
  ".5 eta' (Q+P) eta + g0' eta + const"
  where g0 = -etafix' P
  In conclusion, soft constraints are imposed by modifying the full precision and modifying
  the inner gradient (see "lgc_newton").
  Control this feature with "lgc_setetafix" and "control.penalty".

*/


#include <R_ext/BLAS.h>
#include <R_ext/Lapack.h>
#include <malloc.h>
#ifdef __GNUC__
# undef alloca
# define alloca(x) __builtin_alloca((x))
#elif defined(__sun) || defined(_AIX)
/* this is necessary (and sufficient) for Solaris 10 and AIX 6: */
# include <alloca.h>
#endif
#include "Matrix.h"
# ifdef _OPENMP
#include <omp.h>
# endif
/* allocate n elements of type t */
#define Alloca(n, t)   (t *) malloc( (size_t) ( (n) * sizeof(t) ) )
double DOUBLEMAX = 1e308;
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define CHECK_BIT(var,pos) ((var) & (1<<(pos)))

/* ========================================================================== */
/* Headers which are not exported by "Matrix" */
/* ========================================================================== */
cholmod_dense *cholmod_zeros
(
    /* ---- input ---- */
    size_t nrow,        /* # of rows of matrix */
    size_t ncol,        /* # of columns of matrix */
    int xtype,          /* CHOLMOD_REAL, _COMPLEX, or _ZOMPLEX */
    /* --------------- */
    cholmod_common *Common
) ;
cholmod_dense *cholmod_ones
(
    /* ---- input ---- */
    size_t nrow,        /* # of rows of matrix */
    size_t ncol,        /* # of columns of matrix */
    int xtype,          /* CHOLMOD_REAL, _COMPLEX, or _ZOMPLEX */
    /* --------------- */
    cholmod_common *Common
) ;
cholmod_sparse *cholmod_band
(
    /* ---- input ---- */
    cholmod_sparse *A, /* matrix to extract band matrix from */
    UF_long k1,         /* ignore entries below the k1-st diagonal */
    UF_long k2,         /* ignore entries above the k2-nd diagonal */
    int mode,           /* >0: numerical, 0: pattern, <0: pattern (no diag) */
    /* --------------- */
    cholmod_common *Common
) ;

#ifdef _EXTERNALLIB_
cholmod_factor *cholmod_analyze
(
    /* ---- input ---- */
    cholmod_sparse *A, /* matrix to order and analyze */
    /* --------------- */
    cholmod_common *Common
) ;
#endif

SEXP chm_dense_to_SEXP(CHM_DN a, int dofree, int Rkind, SEXP dn);
CHM_SP lgc_inv_up(CHM_FR Lfac, CHM_SP Sigma, cholmod_common *c);
CHM_SP lgc_inv_super(CHM_FR Lfac, cholmod_common *c, int sorted);
double lgc_trace_matprod(CHM_SP A, CHM_SP B);
int lgc_pointer_transpose(const CHM_SP A, int *pt);
double lgc_trace_matprod2(CHM_SP A, double *x, double *y);
SEXP dgeMatrix_dgeMatrix_crossprod(SEXP, SEXP, SEXP);
SEXP dgeMatrix_crossprod(SEXP, SEXP);
SEXP dgeMatrix_matrix_solve(SEXP, SEXP);
SEXP dpoMatrix_solve(SEXP);

/* ========================================================================== */
/* Data structures for lgc */
/* ========================================================================== */
typedef struct
{
  int trace;             /* Level of printing */
  double tol;            /* gradient tolerance: mgc<tol ===> Newton converged */
  double mgcmax;         /* huge gradient tolerance: mgc>mgcmax ===> Newton dropout (likelihood=NaN) */
  int maxit;             /* Iteration limit */
  int report;            /* Report mode */
  int countflops;        /* Should we perform flopcount? */
  int nrelax[3];         /* Relaxed amalgamation parameters for CHOLMOD */
  double zrelax[3];
  int experimental;      /* The individual bits of this variable is used to switch between 
			    experimental features.
			    BIT0: Use experimental version of supernodal sparse inverse 
			    BIT1: Use ML instead of REML assuming fixed effects are given by 
			          missing data entries. */
  int externallib;       /* Use external cholmod library - (Not the one from Matrix package) */
  double penalty;        /* Penalty used to fix certain components of eta - see "eta_fix". 
                            In the default setting this parameter is not used */
} control_data;          
typedef struct
{
  CHM_SP Q;              /* Sparse precision matrix of dimension n */
  CHM_SP H;              /* Sparse matrix of dimension n.  H = Q+diag(lambda) */
  CHM_FR LQ;             /* Symbolic analysis of Q */
  CHM_FR LH;             /* Symbolic analysis of H (only num values != LQ) */
  CHM_SP N;              /* Observations (counts) - sparseness means missing */
  CHM_SP lambda;         /* exp(eta) for the non-missing entries */
  CHM_DN eta;            /* etahat of latest likelihood evaluation */ 
  CHM_SP iQ;             /* Workspace for inverse subset algorithm */
  int nparms;            /* Length of Qdot */
  double logdetH;
  double logdetQ;
  double value;          /* negative log likelihood */
  int *ptrans;           /* Pointer to transposed entry of iQ */
  int ndata;             /* Number of non-missing data entries */
  int *data_entry;       /* Pointers to non-missing data entries */
  int *iperm;            /* Inverse permutation */
  CHM_DN grad_eta_theta; /* Gradient of eta wrt. theta */
  double value_eta;      /* Likelihood corresponding to the latest prediction eta */
  CHM_DN eta_start;      /* Initial guess on eta */
  CHM_SP eta_fix;        /* Impose constraint on eta (sparseness means free) */
} lgc_data; 

lgc_data lgc;
control_data control;
cholmod_common c;

double flopcount[10]; /* For flopcounting */
void reset_flopcount(){
  for(int i=0;i<10;i++)flopcount[i]=0.0;
}

/* Cleanup an lgc_data-structure */
void lgc_free(lgc_data *lgc, cholmod_common *c){
  M_cholmod_free_sparse(&lgc->Q,c);
  M_cholmod_free_sparse(&lgc->H,c);
  M_cholmod_free_factor(&lgc->LQ,c);
  M_cholmod_free_factor(&lgc->LH,c);
  M_cholmod_free_sparse(&lgc->N,c);
  M_cholmod_free_sparse(&lgc->lambda,c);
  M_cholmod_free_dense(&lgc->eta,c);
  M_cholmod_free_sparse(&lgc->iQ,c);
  //free(lgc->ptrans);
  //free(lgc->iperm);
  M_cholmod_free_dense(&lgc->grad_eta_theta,c);
  M_cholmod_free_dense(&lgc->eta_start,c);
  M_cholmod_free_sparse(&lgc->eta_fix,c);
}

/* ========================================================================== */
/* Copy stuff from R -> lgc C-structure */
/* Note: when converting from R to C via e.g. "AS_CHM_SP" then pointers are  */
/* shared between R and C. Thus we should not worry about freeing workspace  */
/* as this is done by R. */
/* For reverse conversions via e.g. "M_chm_sparse_to_SEXP" then a real copy */
/* is made and it is thus necessary to free the original object if  */
/* allocated by CHOLMOD. */
/* ========================================================================== */
SEXP lgc_setQ(SEXP Qp){
  M_cholmod_free_sparse (&lgc.Q, &c) ; /* Free old copy */
  M_cholmod_free_sparse (&lgc.H, &c) ; /* Free old copy */
  CHM_SP Q=AS_CHM_SP(Qp);
  lgc.Q = M_cholmod_copy(Q, -1 /* symmetric lower */, 1 /*values*/, &c);
  lgc.H = M_cholmod_copy(Q, -1 /* symmetric lower */, 1 /*values*/, &c);
  return R_NilValue;
}
void TEST_VALID_SPARSE_VECTOR(CHM_SP x){
  if(x->ncol != 1)error("Vector must have ncol=1");
  if(x->nrow != lgc.Q->nrow)error("Vector length must match dimension of Q");
}
SEXP lgc_setN(SEXP Np){
  M_cholmod_free_sparse (&lgc.N, &c) ; /* Free old copy */
  CHM_SP N=AS_CHM_SP(Np);
  TEST_VALID_SPARSE_VECTOR(N);
  lgc.N = M_cholmod_copy_sparse(N, &c);
  lgc.lambda = M_cholmod_copy_sparse(N, &c);
  int *p=lgc.N->p;
  lgc.ndata = p[1];
  lgc.data_entry = lgc.N->i;  
  return R_NilValue;
}
SEXP lgc_setetafix(SEXP etafixp){
  M_cholmod_free_sparse (&lgc.eta_fix, &c) ; /* Free old copy */
  CHM_SP etafix=AS_CHM_SP(etafixp);
  TEST_VALID_SPARSE_VECTOR(etafix);
  lgc.eta_fix = M_cholmod_copy_sparse(etafix, &c);
  return R_NilValue;
}
SEXP lgc_setcontrol(SEXP con){
  control.trace = INTEGER(GET_SLOT(con, install("trace")))[0];
  control.tol = REAL(GET_SLOT(con, install("tol")))[0];
  control.mgcmax = REAL(GET_SLOT(con, install("mgcmax")))[0];
  control.maxit = INTEGER(GET_SLOT(con, install("maxit")))[0];
  control.report = INTEGER(GET_SLOT(con, install("report")))[0];
  for(int i=0;i<3;i++){
    control.nrelax[i] = INTEGER(GET_SLOT(con, install("nrelax")))[i];
    control.zrelax[i] = REAL(GET_SLOT(con, install("zrelax")))[i];
  }
  control.countflops = INTEGER(GET_SLOT(con, install("countflops")))[0];
  control.experimental = INTEGER(GET_SLOT(con, install("experimental")))[0];
  control.externallib = INTEGER(GET_SLOT(con, install("externallib")))[0];
  control.penalty = REAL(GET_SLOT(con, install("penalty")))[0];
  return R_NilValue;
}
SEXP lgc_setetastart(SEXP Etastart){
  double *etax=lgc.eta_start->x;
  double *Nx=lgc.N->x;
  int neta=(lgc.eta->nrow)*(lgc.eta->ncol);
  int i,k;
  for(i=0;i<neta;i++)etax[i]=0.0;
  for(k=0;k<lgc.ndata;k++){
    i=lgc.data_entry[k];
    etax[i]=log(Nx[i]+0.1);
  }
  // User given etastart
  if(!isNull(Etastart)){
    if(control.trace)Rprintf("User given etastart \n");
    if(LENGTH(Etastart)!=neta)error("etastart wrong length\n");
    for(i=0;i<neta;i++)etax[i]=REAL(Etastart)[i];
  }
  lgc.value_eta=DOUBLEMAX;
  return R_NilValue;
}
SEXP lgc_cleanup(){
  lgc_free(&lgc,&c);
  return R_NilValue;
}

/* ========================================================================== */
/* Copy stuff from lgc C-structure -> R */
/* ========================================================================== */
SEXP lgc_getQ(){return M_chm_sparse_to_SEXP(lgc.Q, 0 /* Free */ , 0, 0, "", R_NilValue);}
SEXP lgc_getH(){return M_chm_sparse_to_SEXP(lgc.H, 0 /* Free */ , 0, 0, "", R_NilValue);}
SEXP lgc_getL(){return M_chm_factor_to_SEXP(lgc.LQ, 0 /* Free */);}
SEXP lgc_getLQ(){return M_chm_factor_to_SEXP(lgc.LQ, 0 /* Free */);}
SEXP lgc_getN(){return M_chm_sparse_to_SEXP(lgc.N, 0 /* Free */ , 0, 0, "", R_NilValue);}
SEXP lgc_getvalue(){
  CHM_DN x=N_AS_CHM_DN(&lgc.value,1,1);
  double *xx=x->x;
  if(lgc.value==DOUBLEMAX)xx[0]=R_NaN;
  return chm_dense_to_SEXP(x, 0 /* Free */ , 0, R_NilValue);
}
SEXP lgc_geteta(){
  return chm_dense_to_SEXP(lgc.eta, 0 /* Free */ , 0, R_NilValue);
}
SEXP lgc_getiQ(){return M_chm_sparse_to_SEXP(lgc.iQ, 0 /* Free */ , 0, 0, "", R_NilValue);}
SEXP lgc_getgradeta(){
  return chm_dense_to_SEXP(lgc.grad_eta_theta, 0 /* Free */ , 0, R_NilValue);
}
SEXP lgc_getperm(){
  CHM_FR f=lgc.LQ;
  SEXP ans;
  PROTECT(ans = allocVector(INTSXP, f->n));
  Memcpy(INTEGER(ans), (int*)f->Perm, f->n);
  UNPROTECT(1);
  return ans;
}

/* ========================================================================== */
/* Error handler */
/* ========================================================================== */
void lgc_cholmod_error(int status, const char *file, int line, const char *message)
{
  /* From CHOLMOD/Include/cholmod_core.h : ...status values.
     zero means success, negative means a fatal error, positive is a warning.
  */
  warning(("Cholmod warning '%s' at file:%s, line %d"),
	  message, file, line);
}

/* ========================================================================== */
/* Run the symbolic analysis and prepare workspace - only run once !!! */
/* ========================================================================== */
SEXP lgc_symbolic(SEXP Qp, SEXP Perm){
  lgc.value_eta=DOUBLEMAX;
  M_R_cholmod_start(&c);
  /* TODO: function "setcommon" */
  c.nmethods=9; 
  c.supernodal = CHOLMOD_SUPERNODAL;
  c.final_ll = TRUE; /* Important (LDL=FALSE): otherwise lgc_inv_up gives wrong answer.
		        Alternative is to fix "cholmod_factor_to_sparse" */
  /*  Return quickly if not positive definite */
  c.quick_return_if_not_posdef=TRUE;
  c.error_handler=lgc_cholmod_error;
  /* Supernode amalgamation */
  for(int i=0;i<3;i++){
    c.nrelax[i]=control.nrelax[i];
    c.zrelax[i]=control.zrelax[i];
  }
  lgc_free(&lgc,&c);   /* Cleanup all */
  lgc_setQ(Qp);
  lgc.eta = cholmod_ones (lgc.Q->nrow, 1, lgc.Q->xtype, &c) ;
  lgc.eta_start=M_cholmod_copy_dense(lgc.eta,&c);  
  lgc.eta_fix = NULL; /* Default: fix nothing! */
  if(control.trace)Rprintf("Running symbolic analysis \n");
  if(!isNull(Perm)){
    if(control.trace)Rprintf("User permutation \n");
    int *perm=INTEGER(Perm);
    c.nmethods=1; 
    lgc.LQ = M_cholmod_analyze_p(lgc.Q, perm, NULL, 0, &c);
  } else
  if(!control.externallib){
    lgc.LQ = M_cholmod_analyze(lgc.Q, &c); /* get fill-reducing permutation */
  } else {
#ifdef _EXTERNALLIB_
    if(control.trace)Rprintf("Entering externallib \n");
    c.itype=CHOLMOD_INT;
    lgc.LQ = cholmod_analyze(lgc.Q, &c); /* get fill-reducing permutation */
    if(control.trace)Rprintf("cholmod_analyze: status=%d \n",c.status);
#else
    error("Please set externallib to 0L.");
#endif
  }
  if(control.trace)Rprintf("Chosen ordering %d \n", c.selected);
  if(control.trace)Rprintf("Length of supernodal xslot %d \n", lgc.LQ->xsize);
  if(control.trace)Rprintf("Flopcount %f \n", c.fl);
  double nnzL = lgc.LQ->xsize;
  double nnzQ = lgc.Q->nzmax;
  if(control.trace)Rprintf("Fill-in ratio (nnz(L)/nnz(Q)) %f \n", nnzL/nnzQ);
  if(control.trace)Rprintf("Factor xtype %d \n", lgc.LQ->xtype);  
  if(control.trace)Rprintf("Copying factor \n");
  lgc.LH = M_cholmod_copy_factor(lgc.LQ, &c);
  if(control.trace)Rprintf("Done \n");
  return R_NilValue;
}

/* ========================================================================== */
/* LGC Newton iterations and likelihood */
/* ========================================================================== */
#define LGC_NEWTON_DROP_OUT  { lgc.value=DOUBLEMAX; \
                               if(control.trace)Rprintf("NEWTON_DROP_OUT \n"); \
                               M_cholmod_free_dense(&g,&c); \
                               M_cholmod_free_dense(&g0,&c); \
                               M_cholmod_free_dense(&eta,&c); \
                               return R_NilValue; }

SEXP lgc_newton(){
  double mgc, value, mgc2;
  int i, j, k, iter, ok;
  int n = lgc.Q->ncol;
  CHM_DN g = cholmod_ones (n, 1, lgc.Q->xtype, &c) ;
  CHM_DN g0 = cholmod_zeros (n, 1, lgc.Q->xtype, &c);
  CHM_DN deta; // = cholmod_ones (lgc.Q->nrow, 1, lgc.Q->xtype, &c) ;
  double one [2] = {1,0}, zero [2] = {0,0} ;
  double mgc_previous=DOUBLEMAX;
  double *Qx, *Hx, *Nx, *etax, *gx, *detax, *g0x;
  int *Qi, *Hi, *Ni, *Qp, *Hp, *Np;
  /* Sparse matrix pointers */
  Qx = lgc.Q->x; Qi = lgc.Q->i; Qp = lgc.Q->p;
  Hx = lgc.H->x; Hi = lgc.H->i; Hp = lgc.H->p;
  Nx = lgc.N->x; Ni = lgc.N->i; Np = lgc.N->p;
  /* Dense matrix pointers */
  CHM_DN eta=M_cholmod_copy_dense(lgc.eta_start,&c);
  etax = eta->x;
  gx = g->x;
  g0x = g0->x;
  /* First get log-determinant of Q */
  ok = M_cholmod_factorize (lgc.Q, lgc.LQ, &c) ;
  if(!ok | c.status){ /* Q Not positive definite */
    if(control.trace)Rprintf("lgc: CHOLMOD error. ok=%d \n",ok);
    LGC_NEWTON_DROP_OUT;
  }
  lgc.logdetQ = M_chm_factor_ldetL2(lgc.LQ);

  /* =============== Soft constraint:
     1. will change diagonal of lgc.Q (no problem since LQ 
        and logdetQ are now known)
     2. will change g0
 */
  if(lgc.eta_fix != NULL){
    double penalty=control.penalty;
    double *etafixx=lgc.eta_fix->x;
    int *etafixi=lgc.eta_fix->i;
    int *etafixp=lgc.eta_fix->p;
    for(k=etafixp[0];k<etafixp[1];k++){ /* loop through non-missing fixed-entries */
      j=etafixi[k]; /* column Q(:,j) */
      g0x[j] = -penalty*etafixx[k]; /* change gradient */
      i=Qp[j]; /* index of diagonal entry (assuming lower-symmetric storage!) */
      Qx[i] = Qx[i] + penalty;
      Hx[i] = Qx[i];
    }
  }


  for(iter=0;iter<control.maxit;iter++){ /* Start Newton */
    /* ======= Compute gradient and hessian ======= */
    /* Initialize gradient g=g0 */
    memcpy(gx,g0x,n*sizeof(double));
    /* Y=alpha*(A*X)+beta*Y */
    M_cholmod_sdmult (lgc.Q,
		      0,            /* transpose */ 
		      one,          /* alpha */ 
		      one,          /* beta */
		      eta,          /* X */ 
		      g,            /* Y */
		      &c);
    for(k=Np[0];k<Np[1];k++){ /* loop through non-missing data-entries */
      j=Ni[k]; /* column Q(:,j) */
      gx[j] += exp(etax[j]) - Nx[k]; /* Poisson contributions to gradient */
      i=Qp[j]; /* index of diagonal entry (assuming lower-symmetric storage!) */
      Hx[i] = Qx[i] + exp(etax[j]);
    }
    /* ======= Test for convergence ======= */
    mgc=0.0;
    for(k=0;k<n;k++)mgc=MAX(mgc,fabs(gx[k]));
    if(control.trace)Rprintf("mgc %8.1e\n", mgc);
    if((mgc<control.tol) & (iter!=0))break;
    mgc_previous=mgc;
    /* ======= Test for gradient being too large or nan ======= */
    if(!R_FINITE(mgc))LGC_NEWTON_DROP_OUT;
    if(mgc>control.mgcmax)LGC_NEWTON_DROP_OUT;
    /* ======= eta = eta - solve(H,g) ======= */
    M_cholmod_factorize (lgc.H, lgc.LH, &c) ;
    deta = M_cholmod_solve (CHOLMOD_A, lgc.LH, g, &c) ;
    detax = deta->x;
    for(k=0;k<n;k++)etax[k]-=detax[k];
    /* Free */
    M_cholmod_free_dense(&deta,&c);

  } /* End Newton */

  if(control.maxit==0){ /* No iterations - update Cholesky of H */
    /* H=Q+D */
    for(k=Np[0];k<Np[1];k++){ /* loop through non-missing data-entries */
      j=Ni[k]; /* column Q(:,j) */
      i=Qp[j]; /* index of diagonal entry (assuming lower-symmetric storage!) */
      Hx[i] = Qx[i] + exp(etax[j]);
    }
    M_cholmod_factorize (lgc.H, lgc.LH, &c) ;
  }

  /* ========== HACK: to use ML instead of REML 
   * add penalty in order to get logdet(Q+D) instead of logdet(R+D) (see paper) 
   */
  if(CHECK_BIT(control.experimental,1)){
    for(k=0;k<n;k++){
      Hx[Qp[k]] += control.penalty; /* All diagonal entries*/
    }
    for(k=Np[0];k<Np[1];k++){ /* loop through non-missing data-entries */
      j=Ni[k]; /* column Q(:,j) */
      i=Qp[j]; /* index of diagonal entry (assuming lower-symmetric storage!) */
      Hx[i] -= control.penalty; /* Exclude diagonal entries with data */
    }
    // Redo factorization
    M_cholmod_factorize (lgc.H, lgc.LH, &c) ;
  }
  /* end hack... */


  lgc.logdetH = M_chm_factor_ldetL2(lgc.LH);

  /* ======= Compute likelihood ======= 
    The Laplace approximation is given by  (see paper): 
    value = sum(exp(eta)) - sum(N*eta) + .5*t(eta)*Q*eta - .5*logdetQ + .5*logdetH
    Note there is a simple formula for the quadratic form because gradient=0:
    .5*t(eta)*Q*eta = .5*( t(eta)*(N-exp(eta)) )
    Insert this in the expression for "value":
    [ Note change: .5*t(eta)*Q*eta + t(eta)*g0 = .5*( t(eta)*(N-exp(eta)) ) + .5*t(eta)*g0 ]
  */
  value=0.0;
  for(k=Np[0];k<Np[1];k++){ /* loop through non-missing data-entries */
    j=Ni[k]; /* column Q(:,j) */
    value += exp(etax[j]) - Nx[k]*etax[j] + .5*etax[j]*(Nx[k]-exp(etax[j]));
  }
  /* Contribution from linear term sum(g0*eta) */
  for(k=0;k<n;k++)value+=.5*etax[k]*g0x[k]; /* see note above for why ".5" */

  /* quadratic form directly: value += .5*lgc_trace_matprod2(lgc.Q,eta,eta); */
  value += -.5*lgc.logdetQ + .5*lgc.logdetH;
  
  lgc.value=value;

  /* update etahat */
  M_cholmod_free_dense(&lgc.eta,&c);
  lgc.eta = eta;

  /* Update eta_start with *copy* of etahat if likelihood improved */
  if(value<lgc.value_eta){
    M_cholmod_free_dense(&lgc.eta_start,&c);
    lgc.eta_start = M_cholmod_copy_dense(eta,&c);
    lgc.value_eta=value;
  } 

  M_cholmod_free_dense(&g,&c);
  M_cholmod_free_dense(&g0,&c);
    
  return R_NilValue;
}

/* ========================================================================== */
/* LGC gradient of likelihood - A prior call to lgc_newton is required !!! */
/* ========================================================================== */
SEXP lgc_gradient(SEXP Qdotlist){
  /* Update eta_start when computing the gradient */
  if(FALSE){
    M_cholmod_free_dense(&lgc.eta_start,&c);
    lgc.eta_start = M_cholmod_copy_dense(lgc.eta,&c);
  }

  int n=lgc.Q->ncol;
  int nparms=LENGTH(Qdotlist);
  CHM_SP Qdot[nparms];
  CHM_DN ans = cholmod_zeros (nparms, 1, lgc.Q->xtype, &c);
  double *ansx=ans->x;
  double *etax=lgc.eta->x;
  int *dataindex=lgc.N->i;
  int ndata=lgc.N->nrow;
  int i;
  /* TODO: Check that Qdotlist contains "dgCMatrix" */
  for(i=0;i<nparms;i++)Qdot[i]=AS_CHM_SP(VECTOR_ELT(Qdotlist,i));
  
  //for(i=0;i<nparms;i++)ansx[i]=0;
  /* Apply subset of inverse on Q */

  /* SUPERNODAL */
  //CHM_SP iQ = lgc_inv_up(lgc.LQ, lgc.iQ, &c);
  CHM_SP iQ = lgc_inv_super(lgc.LQ, &c, FALSE /*sorted*/ );

#pragma omp parallel for
  for(i=0;i<nparms;i++)ansx[i]-=.5*lgc_trace_matprod(Qdot[i],iQ);
  M_cholmod_free_sparse(&iQ,&c);

 
  /* Apply subset of inverse on H=Q+D */
  /* SUPERNODAL */
  //CHM_SP iH = lgc_inv_up(lgc.LH, lgc.iQ, &c);
  CHM_SP iH = lgc_inv_super(lgc.LH, &c, FALSE /*sorted*/ );

#pragma omp parallel for
  for(i=0;i<nparms;i++)ansx[i]+=.5*lgc_trace_matprod(Qdot[i],iH);
  /* Term .5*xi^T*R*xi */
#pragma omp parallel for
  for(i=0;i<nparms;i++)ansx[i]+=.5*lgc_trace_matprod2(Qdot[i],etax,etax);
  /* Term -t(v)*H^-1*Rdot*xi where v=diag(H)*lambda */
  CHM_SP diagH = cholmod_band(iH,0,0,1,&c);
  CHM_DN vec = cholmod_zeros (n, 1, iH->xtype, &c) ;
  double *diagHx=diagH->x;
  double *vecx=vec->x;
  for(int k=0;k<lgc.ndata;k++){
    i=lgc.data_entry[k];
    vecx[i]=exp(etax[i])*diagHx[i];
  }
  CHM_DN vec2 = M_cholmod_solve (CHOLMOD_A, lgc.LH, vec, &c);

#pragma omp parallel for
  for(i=0;i<nparms;i++)ansx[i]-=.5*lgc_trace_matprod2(Qdot[i],vec2->x,etax);

  /* If we are in report-mode also compute grad_eta_theta = -H^-1*Rdot*xi */
  if(control.report){
    double minus_one [2] = {-1,0}, zero [2] = {0,0} ;
    M_cholmod_free_dense(&(lgc.grad_eta_theta),&c);
    lgc.grad_eta_theta = M_cholmod_allocate_dense(n,nparms,n,CHOLMOD_REAL,&c);
    CHM_DN vec3 = M_cholmod_copy_dense(lgc.eta,&c);
    CHM_DN vec4;
    double *dest=lgc.grad_eta_theta->x;
    for(i=0;i<nparms;i++){
      /* Y=alpha*(A*X)+beta*Y */
      M_cholmod_sdmult (Qdot[i],
			0,            /* transpose */ 
			minus_one,    /* alpha */ 
			zero,         /* beta */
			lgc.eta,      /* X */ 
			vec3,         /* Y */
			&c);
      vec4 = M_cholmod_solve (CHOLMOD_A, lgc.LH, vec3, &c);
      memcpy(dest,vec4->x,n*sizeof(double));
      dest+=n;
      M_cholmod_free_dense(&vec4,&c);
    }
    M_cholmod_free_dense(&vec3,&c);
  }

  M_cholmod_free_dense(&vec,&c);
  M_cholmod_free_dense(&vec2,&c);
  M_cholmod_free_sparse(&iH,&c);
  M_cholmod_free_sparse(&diagH,&c);

  return chm_dense_to_SEXP(ans, 1 /* cholmod_free */ , 0, R_NilValue);

}

/* =========== To check the routines ======== */
SEXP lgc_factorize(){
  M_cholmod_factorize(lgc.Q, lgc.LQ, &c);
  return R_NilValue;
}
SEXP lgc_solvesubset(){
  CHM_SP iQ = lgc_inv_up(lgc.LQ, lgc.iQ, &c);
  M_cholmod_free_sparse(&iQ,&c);
  return R_NilValue;
}

/* Inverse subset of H=Q+D
   A Prior call to "lgc_newton" is required. */
SEXP lgc_invH(){
  CHM_SP iH = lgc_inv_super(lgc.LH, &c, TRUE /* sorted */);
  return M_chm_sparse_to_SEXP(iH, 1 /* Free */ , 0, 0, "", R_NilValue);
}

SEXP lgc_experimental(){
  int x=control.experimental;
  for(int i=0;i<16;i++){
    //Rprintf("Bits of experimental: %d \n", CHECK_BIT(x,i));
    Rprintf("Bit: %d ", i);
    if(CHECK_BIT(x,i)) Rprintf("*");
    Rprintf("\n");
  }
  return R_NilValue;
}

/* sample from N(0,Q^-1) 
   Input: matrix of iid N(0,1)
   Note: Assuming LL (NOT LDL!)
*/
SEXP lgc_GMRFsample(SEXP x){
  CHM_DN u=AS_CHM_DN(x);
  int ok = (u->nrow == lgc.LQ->n);
  if(!ok)error("Matrix dimension must match factor dimension.");
  CHM_DN u1=M_cholmod_solve(CHOLMOD_Lt,lgc.LQ,u,&c);
  CHM_DN u2=M_cholmod_solve(CHOLMOD_Pt,lgc.LQ,u1,&c);
  M_cholmod_free_dense(&u1,&c);
  return chm_dense_to_SEXP(u2, 1 /* Free */ , 0, R_NilValue);
}
/* Same for N(0,H^-1) */
SEXP lgc_GMRFsampleH(SEXP x){
  CHM_DN u=AS_CHM_DN(x);
  int ok = (u->nrow == lgc.LH->n);
  if(!ok)error("Matrix dimension must match factor dimension.");
  CHM_DN u1=M_cholmod_solve(CHOLMOD_Lt,lgc.LH,u,&c);
  CHM_DN u2=M_cholmod_solve(CHOLMOD_Pt,lgc.LH,u1,&c);
  M_cholmod_free_dense(&u1,&c);
  return chm_dense_to_SEXP(u2, 1 /* Free */ , 0, R_NilValue);
}


/* Solve system solve(H)%*%x 
   Note: Assuming factorization lgc.LH exists in workspace
*/
SEXP lgc_Hsolve(SEXP x){
  CHM_DN u=AS_CHM_DN(x);
  int ok = (u->nrow == lgc.LH->n);
  if(!ok)error("Matrix dimension must match factor dimension.");
  CHM_DN u2 = M_cholmod_solve (CHOLMOD_A, lgc.LH, u, &c) ;
  return chm_dense_to_SEXP(u2, 1 /* Free */ , 0, R_NilValue);
}

/* ========================================================================== */
/* What follows is copied from chol.c and converted to CHOLMOD structures */
/* ========================================================================== */

/* Inner product of sparse matrix-columns */
/* < A(:,i) , B(:,j) >  */
/* Assumes *sorted* row indices  */
/* (can be obtained by double transposing the matrices) */
double lgc_inner_prod(CHM_SP A, CHM_SP B, int i, int j){
  int *Ap=A->p, *Bp=B->p, *Ai=A->i, *Bi=B->i, pa, pb;
  double *Ax=A->x, *Bx=B->x, sum=0.0;
  pb=Bp[j];
  for(pa=Ap[i];pa<Ap[i+1];pa++){
    while( Bi[pb]<Ai[pa] & pb<Bp[j+1] )pb++;
    if( Bi[pb]==Ai[pa] & pb<Bp[j+1] )sum+=Ax[pa]*Bx[pb];
  }
  return sum;
}
/* <dense,sparse> inner product  < A(:) , B(:,j) > */
/* (so that length of A must be B->m) */
/* j=column of B to consider */
/* pstart=pointer to nonzero element of B we want to start with */
/* (if pstart=B->p[j] its just an ordinary inner product)  */
double lgc_inner_prod3(double *A, CHM_SP B, int j, int p_start){
  int *Bp=B->p, *Bi=B->i, p;
  double *Bx=B->x, sum=0.0;
  for(p=p_start;p<Bp[j+1];p++){
    sum+=Bx[p]*A[Bi[p]];
  }
  return sum;
}
/* stolen cs_cumsum from cs.c */
double lgc_cumsum (int *p, int *c, int n)
{
    int i, nz = 0 ;
    double nz2 = 0 ;
    if (!p || !c) return (-1) ;     /* check inputs */
    for (i = 0 ; i < n ; i++)
    {
        p [i] = nz ;
        nz += c [i] ;
        nz2 += c [i] ;              /* also in double to avoid int overflow */
        c [i] = p [i] ;             /* also copy p[0..n-1] back into c[0..n-1]*/
    }
    p [n] = nz ;
    return (nz2) ;                  /* return sum (c [0..n-1]) */
}
/* Find pointer to transpose entry */
int lgc_pointer_transpose(const CHM_SP A, int *pt){
  int *w,*w2,p,m,n,*Ap,*Ai,j;
  m=A->nrow; n=A->ncol; Ap=A->p; Ai=A->i;
  w=Alloca(m,int);  /* initialized to zero */
  w2=Alloca(m+1,int);  /* dummy workspace */
  for(p=0;p<m;p++){w[p]=0;w2[p]=0;}
  w2[m]=0;
  for(p=0;p<Ap[n];p++)w[Ai[p]]++; /* row counts */
  lgc_cumsum(w2,w,m); /* row pointers */
  for(j=0;j<n;j++){
    for(p=Ap[j];p<Ap[j+1];p++){
      pt[p]=w[Ai[p]]++;
    }
  }
  //cs_free(w); cs_free(w2);
  free(w); free(w2);
  return 0;
}
/* Trace of sparse matrix-product tr(AB) for *symmetric* A and B. */
/* A and B are assumed to be in full storage (dgCMatrix) */
/* Assumes *sorted* row indices */
double lgc_trace_matprod_old(CHM_SP A, CHM_SP B){
  double sum=0.0;
  int ncol=A->ncol;
  /* TODO: OPENMP parallelize ? */
  for(int i=0;i<ncol;i++) sum+=lgc_inner_prod(A,B,i,i);
  return sum;
}

/* Trace of matrix-product tr(t(x)*A*y) for dense vectors x,y and sparse A */
/* A: symmetric sparse matrix in *asymmetric* storage!  */
double lgc_trace_matprod2(CHM_SP A, double *x, double *y){
  double sum=0.0;
  int ncol=A->ncol, stype=A->stype;
  double *Ax;
  int *Ai, *Ap, i;
  Ai=A->i; Ap=A->p; Ax=A->x;
  /* TODO: OPENMP parallelize ? */
  for(int j=0;j<ncol;j++){
    for(int k=Ap[j];k<Ap[j+1];k++){
      i=Ai[k];
      if(stype==0){ /* unsym storage */
	sum+=Ax[k]*x[i]*y[j];
      } else { /* Upper or Lower triangular storage */
	sum+=(i==j ? Ax[k]*x[i]*y[j] : 2*Ax[k]*x[i]*y[j] );
      }
    }
  }
  return sum;
}

/*
  Trace of sparse matrix-product tr(AB) *symmetric* A and B.
  A: full asymmetric storage.
  B: triangular storage.
  row indices may be *unsorted*
*/
double lgc_trace_matprod(CHM_SP A, CHM_SP B){
  double sum=0.0;
  int ncol=A->ncol,i,j;
  double *w=Alloca(ncol,double);
  double *Ax=A->x, *Bx=B->x;
  int *Ai=A->i, *Bi=B->i, *Ap=A->p, *Bp=B->p;
  for(i=0;i<ncol;i++)w[i]=0.0;
  for(j=0;j<ncol;j++){
    for(i=Ap[j];i<Ap[j+1];i++)w[Ai[i]]=Ax[i]; /* scatter */
    /* Works for both lower and upper (or mixture - permutation of triangular) */
    /* TODO: branch according to B->stype */
    for(i=Bp[j];i<Bp[j+1];i++)sum+= (Bi[i]==j ? w[Bi[i]]*Bx[i] : 2*w[Bi[i]]*Bx[i]);
    for(i=Ap[j];i<Ap[j+1];i++)w[Ai[i]]=0.0; /* clear */
  }   
  free(w);
  return sum;
}

/* ======= Inverse Subset Algorithm ====== */
/* Recursion: Compute Sigma(:,k) assuming known entries below diagonal */
/* Sigma(i,k) = 1_(k==i)/(Lii*Lii) - 1/Lii*sum(L0(:,i)*Sigma(:,k))  */
/* FLOPCOUNT: The number of multiply-add instructions required to */
/* compute k'th column of Sigma is L*(L^T*1) where "L" denotes */
/* the *pattern* of the lower Cholesky and "1" is the dense */
/* vector of ones. */
/* The total flop-count required by the entire algorithm is  */
/* thus 1^T*L*(L^T*1) = sum(colcounts^2) where "colcounts" */
/* denotes the column-counts of L. */
void lgc_recursion(CHM_SP L, CHM_SP Sigma, int k, int *ptrans, int delay, int* col_done){
  int i,n=Sigma->ncol;
  int *Sp=Sigma->p, *Lp=L->p, *Si=Sigma->i, *Li=L->i, ps, pdiag, ptop=Sp[k];
  double *Sx=Sigma->x, *Lx=L->x, Lii;

  /* Workspace to hold *dense* Sigma column. *Must* be cleared */
  double *S = Alloca(n, double);
  for(i=0;i<n;i++)S[i]=0.0;

  /* If "delay" is non-zero we have to wait until col k+1 is done.
     Note: we need an extra flush to make sure we have the updated col_done 
     *and* the updated values of Sigma */

  if(delay){
    //printf("col %d waiting\n",k);
    while(col_done[0]!=(k+1)){
#pragma omp flush(col_done)
    }
#pragma omp flush(col_done,Sx)
    //printf("col %d done waiting\n",k);
  }

  /* Scatter Sigma column - and find diagonal pointer */
  for(ps=Sp[k];ps<Sp[k+1];ps++){
    S[Si[ps]]=Sx[ps];
    if(Si[ps]==k)pdiag=ps;
  }

  /* If "delay" is non-zero we must prepare data for next thread
     and make that data visible */
  if(delay){  
    for(int p=pdiag;p>=ptop;p--){
      i=Si[p];
      if(i<k-delay)break;
      Lii=Lx[Lp[i]];
      S[i]=-1/Lii*lgc_inner_prod3(S,L,i,Lp[i]+1);
      if(k==i)S[i]+=1.0/(Lii*Lii);
      Sx[ptrans[p]]=S[i];
    }
    col_done[0]=k;
#pragma omp flush(col_done,Sx)
  }

  /* Compute dense column */
  for(int p=pdiag;p>=ptop;p--){
    i=Si[p];
    Lii=Lx[Lp[i]];
    S[i]=-1/Lii*lgc_inner_prod3(S,L,i,Lp[i]+1);
    if(k==i)S[i]+=1.0/(Lii*Lii);
  }

  /* Write to sparse and transposed entries */
  for(int p=pdiag;p>=ptop;p--){
    i=Si[p];
    Sx[p]=S[i];
    Sx[ptrans[p]]=S[i];
  }

  /* Now Sx is updated and future threads must see the changes */
  if(delay){
#pragma omp flush(Sx)
    //printf("completed col %d\n",k);
  }

  //cs_free(S);
  free(S);
}

/* Remember: Sigma holds L+t(L) pattern with *sorted* storage */
/* Call: cholmod_sort at initialization */
/* lgc_symbolic should do this: */
/*   /\* Compute sparsity pattern of (subset of) Sigma: (L+LT) *\/ */
/*   cs* LT = cs_transpose (L, 1) ;	/\* compute L' *\/ */
/*   cs *Sigma,*tmp; */
/*   tmp=cs_add(L,LT,0.0,0.0); */
/*   Sigma=cs_transpose(tmp,1); /\* Transpose symmetric matrix to sort columns *\/ */
/*   cs_spfree(tmp);cs_spfree(LT); */


/* Input: Factor L and workspace Sigma = L+t(L) */
CHM_SP lgc_inv_up(CHM_FR Lfac, CHM_SP Sigma, cholmod_common *c){

  /* Convert factor to sparse without modifying factor */
  CHM_FR Ltmp = M_cholmod_copy_factor(Lfac,c);
  CHM_SP L = M_cholmod_factor_to_sparse(Ltmp,c);
  M_cholmod_free_factor(&Ltmp,c);

  int n = L->ncol;
  int *iperm = Alloca(n,int);
  int *perm = Lfac->Perm;
  for(int i=0;i<n;i++)iperm[perm[i]]=i;

  /* Permutation to map p'th element of Sigma to transposed entry  */
  //int *ptrans = Alloca(Sigma->nzmax, int);
  //lgc_pointer_transpose(Sigma,ptrans);
  int *ptrans=lgc.ptrans;

  /* Parallel column version  */

  int col_done, i;
  int delay=0;
# ifdef _OPENMP
  delay=omp_get_max_threads();
# endif

  col_done=n;
#pragma omp parallel for schedule(static,1)  
  for(i=n-1;i>=0;i--)lgc_recursion(L,Sigma,i,ptrans,delay,&col_done);

  /* Serial version */
/*   for(i=n-1;i>=0;i--){ */
/*     recursion(L,Sigma,i,ptrans,0,&col_done); */
/*   } */

  /* Apply inverse permutation on inverse subset */
  M_cholmod_free_sparse(&L,c);
  //CHM_SP ans = cholmod_ptranspose(Sigma,1,iperm,iperm,n,c);
  //M_cholmod_sort(ans,c);
 
  //return M_cholmod_submatrix (Sigma,iperm,n,iperm,n,TRUE,TRUE,c) ;
  CHM_SP ans=M_cholmod_submatrix (Sigma,iperm,n,iperm,n,TRUE,TRUE,c) ;
  free(iperm);
  return ans;
  
}


/* ===========================================================================  */
/* =========== Super-nodal version of solvesubset */


/* According to the documentation of CHOLMOD we have: */
/*         k1 = Super [s] ;            /\* s contains columns k1 to k2-1 of L *\/ */
/*         k2 = Super [s+1] ; */
/*         nscol = k2 - k1 ;           /\* # of columns in all of s *\/ */
/*         psi = Lpi [s] ;             /\* pointer to first row of s in Ls *\/ */
/*         psx = Lpx [s] ;             /\* pointer to first row of s in Lx *\/ */
/*         psend = Lpi [s+1] ;         /\* pointer just past last row of s in Ls *\/ */
/*         nsrow = psend - psi ;       /\* # of rows in all of s *\/ */
/* ============ See discussion in cholmod_change_factor.c */
/*  * (3) supernodal symbolic:  A representation of the nonzero pattern of the */
/*  *      supernodes for a supernodal factorization.  There are L->nsuper */
/*  *      supernodes.  Columns L->super [k] to L->super [k+1]-1 are in the kth */
/*  *      supernode.  The row indices for the kth supernode are in */
/*  *      L->s [L->pi [k] ... L->pi [k+1]-1].  The numerical values are not */
/*  *      allocated (L->x), but when they are they will be located in */
/*  *      L->x [L->px [k] ... L->px [k+1]-1], and the L->px array is defined */
/*  *      in this factor type. */



/* Dense subset x[p,q] of sparse matrix x */
/* If the nz-pattern of x[p,q] is completely dense use "reset=FALSE" */
/* Otherwise use "reset=TRUE" */
/* If p==q and x is symmetric with lower triangle storage */
/* use "symm=TRUE" to get symmetric output. */
CHM_DN densesubmatrix(CHM_SP x, int *p, int np, int *q, int nq, int reset, int symm){
  CHM_DN ans = M_cholmod_allocate_dense(np,nq,np,CHOLMOD_REAL,&c);
  double *w = Alloca(x->nrow,double);
  int *xi=x->i;
  int *xp=x->p;
  double *xx=x->x;
  double *ansx=ans->x;
  int col, row;
  for(int j=0;j<nq;j++){
    col=q[j];
    /* reset workspace w[p] on request */
    if(reset){
      for(int i=0;i<np;i++)w[p[i]]=0;
    }
    /* scatter x[:,col] over w */
    for(int i=xp[col];i<xp[col+1];i++){
      w[xi[i]]=xx[i];
    }
    /* Copy w[p] to ans[:,col] */
    //for(int i=0;i<np;i++){
    for(int i=j;i<np;i++){
      row=p[i];
      
      if(FALSE){ /* old code */
	ansx[0]=w[row];
	ansx++;
      }
      
      ansx[i+j*np]=w[row];
      if(symm)ansx[j+i*np]=w[row];

    }
  }
  free(w);
  return ans;
}

/* Perform recursions for k'th supernode */
void lgc_recursion_super(CHM_SP Lsparse, int k){
  int EXPERIMENTAL = CHECK_BIT(control.experimental,0);

  /* Input */
  CHM_FR L=lgc.LQ;

  int* super=L->super;
  int* Ls=L->s;
  int* Lpi=L->pi;

  int ncol=super[k+1]-super[k]; /* ncol of supernode */
  int nrow=Lpi[k+1]-Lpi[k]; /* Number of rows in supernode */

  /* q contains row-indices of *entire* supernode */
  /* p contains row-indices excluding those of diagonal */
  /* s contains row-indices of diagonal - setdiff(q,p) */
  int* q=Ls+Lpi[k];    /* Pointer to L->s [L->pi [k]] */
  int nq=nrow;         /* length of q */
  int* p=q+ncol;       /* Exclude triangle in diagonal */
  int np=nq-ncol;      /* length of p */
  int* s=q;
  int ns=ncol;         /* length of s */
  /* do not sort because p is sorted */

  int info; /* For lapack */
  int i,j;

  /* NEW ALGORITHM */
  double ONE=1.0, ZERO=0.0, MONE=-1.0;
  CHM_DN x = densesubmatrix(Lsparse,q,nq,q,nq,FALSE,EXPERIMENTAL);
  double *xx=x->x;
  double *Lss=xx, *Lps=xx+ns, *Ssp=xx+(nq*ns), *Spp=xx+(nq*ns+ns);

  /* Workspace to hold output from dsymm */
  double *wrk=malloc((nq*ns)*sizeof(double));
  double *wrkps=wrk+ns;

  if(!EXPERIMENTAL){
  /* ------------ ORIGINAL VERSION: S(p,p)*M */
  if(np>0){
    F77_CALL(dtrsm)("Right","Lower","No transpose","Not unit",
		    &np,&ns,&MONE,Lss,&nq,Lps,&nq);
    for(i=ns;i<nq;i++){for(j=0;j<ns;j++)Lss[j+nq*i] = Lss[i+nq*j];} /* Copy Transpose*/
    //OLD: F77_CALL(dsymm)("Left","Lower",&np,&ns,&ONE,Spp,&nq,Lps,&nq,&ZERO,Lps,&nq);
    memcpy(wrk,Lss,nq*ns*sizeof(double));    
    F77_CALL(dsymm)("Left","Lower",&np,&ns,&ONE,Spp,&nq,Lps,&nq,&ZERO,wrkps,&nq);
    memcpy(Lss,wrk,nq*ns*sizeof(double));
    F77_CALL(dpotri)("L",&ns,Lss,&nq,&info);
    F77_CALL(dgemm)("N","N",&ns,&ns,&np,&ONE,Ssp,&nq,Lps,&nq,&ONE,Lss,&nq);
  } else {
    F77_CALL(dpotri)("L",&ns,Lss,&nq,&info);
  }


  /* ----------- Fill results into L(q,s) -------*/
  double *Lx=Lsparse->x;
  int *Lp=Lsparse->p;
  int m=Lp[s[0]];
  for(j=0;j<ns;j++){
    for(i=j;i<nq;i++){
      Lx[m]=Lss[i+j*nq];
      m++;
    }
  }

  /* ------------ END: ORIGINAL VERSION: S(p,p)*M */
  }


  if(EXPERIMENTAL){
  /* ------------ EXPERIMENTAL VERSION: M^t * S(p,p) */
  /* NB: REQUIRES "symm=TRUE" in densesubmatrix */
  if(np>0){
    F77_CALL(dtrsm)("Right","Lower","No transpose","Not unit",
		    &np,&ns,&MONE,Lss,&nq,Lps,&nq);
    F77_CALL(dgemm)("T","N",&ns,&np,&np,&ONE,Lps,&nq,Spp,&nq,&ZERO,Ssp,&nq);
    F77_CALL(dpotri)("U",&ns,Lss,&nq,&info);
    F77_CALL(dgemm)("N","N",&ns,&ns,&np,&ONE,Ssp,&nq,Lps,&nq,&ONE,Lss,&nq);
  } else {
    F77_CALL(dpotri)("U",&ns,Lss,&nq,&info);
  }


  /* ----------- Fill results into L(q,s) -------*/
  double *Lx=Lsparse->x;
  int *Lp=Lsparse->p;
  int m=Lp[s[0]];
  for(i=0;i<ns;i++){
    for(j=i;j<nq;j++){ 
      Lx[m]=Lss[i+j*nq];
      m++;
    }
  }

  /* ------------ END: EXPERIMENTAL VERSION: M^t * S(p,p) */
  }


  /* Count flops */
  if(control.countflops){
    double Ns=ns;
    double Np=np;
    if(np>0){
      flopcount[0]+=(Ns*(Ns+1))*0.5*Np; /* dtrsm */
      flopcount[1]+=Np*Np*Ns; /* dsymm */
      flopcount[2]+=2.0*(Ns*Ns*Ns)/3.0; /* dpotri */
      flopcount[3]+=Ns*Np*Ns; /* dgemm */
    } else {
      flopcount[2]+=2.0*(Ns*Ns*Ns)/3.0; /* dpotri */
    }
  }


  /* Clean up */
  M_cholmod_free_dense(&x,&c);
  free(wrk);

}

CHM_SP lgc_inv_super(CHM_FR Lfac, cholmod_common *c, int sorted){
  /* Convert factor to sparse without modifying factor */
  CHM_FR Ltmp = M_cholmod_copy_factor(Lfac,c);
  CHM_SP L = M_cholmod_factor_to_sparse(Ltmp,c);
  M_cholmod_free_factor(&Ltmp,c);

  /* inverse permutation */
  int n = L->ncol;
  int *iperm = Alloca(n,int);
  int *perm = Lfac->Perm;
  for(int i=0;i<n;i++)iperm[perm[i]]=i;

  /* Loop over supernodes */
  int nsuper=Lfac->nsuper;
  for(int k=nsuper-1;k>=0;k--)lgc_recursion_super(L,k);

  /* Apply inverse permutation on inverse subset */
  /* WE NO LONGER ASSUME SORTED ROW INDICES */
  /* HOWEVER: FOR COMPARISON WITH OLD METHOD USE SORTED=TRUE  */
  CHM_SP ans=M_cholmod_submatrix (L,iperm,n,iperm,n,TRUE,sorted /*sorted?*/,c) ;
  free(iperm);
  M_cholmod_free_sparse(&L,c);
  return ans;
}
